/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) <year> <name of author>
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.graphview;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;

import javax.swing.JTabbedPane;

import belvedere.graphview.*;
import belvedere.model.*;
import belvedere.util.ButtonDescriptor;
import belvedere.util.DropTarget;

/**
 * This class creates the JInternalFrame window that holds the various views of
 * a model.
 */
public class GraphViewPanel extends JPanel implements KeyListener, ModelChangeListener,
                                                  DropTarget, ActionListener {

  private Model mdl; // the Model held by this GraphViewPanel

  // ArrayLists to hold the buttons for Nodes/Relations loaded from the Model
  // two different array lists so we know which state we are in
  private ArrayList addNodeButtons = new ArrayList();
  private ArrayList addRelationButtons = new ArrayList();

  private JButton btnDelete;
  private GraphPanel pnlGraph;
  private JTextField txtTextEditor;

  /**
   * public constructor to create a ModelFrame object
   * 
   * @param title  the string to put in the title of the ModelFrame window
   */
  public GraphViewPanel(Model mdl, JFrame parent) {
    super();
    //for now we just create an empty new model, later the model is probably
    //passed to this ModelFrame window
    this.mdl = mdl;
    setLayout(new BorderLayout());
    //set up the toolbar
    JToolBar tb = new JToolBar();
    tb.setFloatable(false); // floating toolbar can disappear behind the app.
    tbAddButtons(tb);
    add(BorderLayout.NORTH,tb);
    //set up the main pane with 5 pixel gap in the vertical
    JPanel pnlMain = new JPanel(new BorderLayout(0,5));
    //the text field for edits/new
    this.txtTextEditor = new JTextField();
    this.txtTextEditor.addKeyListener(this);
    pnlMain.add(BorderLayout.NORTH, txtTextEditor);
    this.pnlGraph = new GraphPanel(mdl,parent);
    this.mdl.addModelChangeListener(pnlGraph);
    pnlMain.add(BorderLayout.CENTER,new JScrollPane(pnlGraph,
                                                         JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
                                                         JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS));
    add(BorderLayout.CENTER,pnlMain);
  }

  /**
   * configures the given toolbar
   * 
   * @param toolbar the toolbar to configure
   */
  protected void tbAddButtons(JToolBar toolbar) {
    Rectangle bounds = new Rectangle();
    //Bounds for each button
    bounds.x=2;
    bounds.y=2;
    bounds.width=40;
    bounds.height=40;
    //Toolbar separator
    Dimension separator = new Dimension(5,5);
    Dimension buttonsize = new Dimension(bounds.width,bounds.height);

    //add the ArrayList of Node adding buttons returned by the Model
    ArrayList buttonArray = mdl.getNodeButtonDescriptors();
    ButtonDescriptor btnDescriptor;
    JButton btnAddNode;
    ImageIcon icon;
    for (int i=0;i<buttonArray.size();i++) {
      btnDescriptor = (ButtonDescriptor)buttonArray.get(i);
      icon = new ImageIcon(btnDescriptor.getImageIcon());
      btnAddNode = new JButton(icon);
      btnAddNode.setActionCommand(btnDescriptor.getActionCommand());
      btnAddNode.setToolTipText(btnDescriptor.getToolTipText());
      btnAddNode.addActionListener(this);
      //btnAddNode.setMnemonic();
      toolbar.add(btnAddNode);
      btnAddNode.setBounds(bounds);
      btnAddNode.setMinimumSize(buttonsize);
      btnAddNode.setMaximumSize(buttonsize);
      btnAddNode.setPreferredSize(buttonsize);
      addNodeButtons.add(btnAddNode);
    }

    // Separator
    toolbar.addSeparator(separator);

    //add the ArrayList of Relation adding buttons returned by the Model
    buttonArray = mdl.getRelationButtonDescriptors();
    JToggleButton btnAddRelation;
    for (int i=0;i<buttonArray.size();i++) {
      btnDescriptor = (ButtonDescriptor)buttonArray.get(i);
      icon = new ImageIcon(btnDescriptor.getImageIcon());
      btnAddRelation = new JToggleButton(icon);
      btnAddRelation.setActionCommand(btnDescriptor.getActionCommand());
      btnAddRelation.setToolTipText(btnDescriptor.getToolTipText());
      btnAddRelation.addActionListener(this);
      //btnAddRelation.setMnemonic();
      toolbar.add(btnAddRelation);
      bounds.x += bounds.width;
      btnAddRelation.setBounds(bounds);
      btnAddRelation.setMinimumSize(buttonsize);
      btnAddRelation.setMaximumSize(buttonsize);
      btnAddRelation.setPreferredSize(buttonsize);
      addRelationButtons.add(btnAddRelation);
    }

    // Separator
    toolbar.addSeparator(separator);
    //DeleteNode
    icon   = new ImageIcon("../images/delete.gif");
    btnDelete = new JButton(icon);
    btnDelete.setDefaultCapable(false);
    btnDelete.setEnabled(false);
    btnDelete.setToolTipText("Delete");
    btnDelete.addActionListener(this);
    //btnDelete.setMnemonic((int)'N');
    toolbar.add(btnDelete);
    btnDelete.setBounds(bounds);
    btnDelete.setMinimumSize(buttonsize);
    btnDelete.setMaximumSize(buttonsize);
    btnDelete.setPreferredSize(buttonsize);
    //this control takes all the space on the right, forcing buttons left
    toolbar.add(Box.createHorizontalGlue());
  }

  // currently unused KeyListener events
  public void keyTyped(KeyEvent e) {}
  public void keyPressed(KeyEvent e) {}

  /**
   * executed when the key is released within the textEditor
   * 
   * @param e      the KeyEvent passed to this method
   */
  public void keyReleased(KeyEvent e) {
    Constituent selected = mdl.getSelection();
    if (selected!=null) {
      // the reason the model sets the text instead of doing it directly
      // is so that the associated ModelChangeListener event can fire
      this.mdl.setConstituentText(selected,txtTextEditor.getText());
      this.mdl.setDirty(true);
    } else if (e.getKeyCode()==KeyEvent.VK_ENTER) {
      // models should probably implement a "default" Node
      // pnlGraph.addNode("default",txtTextEditor.getText());
    }
  }

  /**
   * ActionListener event for responding to button clicks
   * 
   * @param event  associated ActionEvent
   */
  public void actionPerformed(ActionEvent event) {
    Object object = event.getSource();
    if (object==btnDelete) {
      // constituentSelect makes sure this button is enabled only
      // when something is selected
      this.mdl.deleteSelected();
    } else {
      if (object instanceof JButton)
        this.pnlGraph.addNode(((JButton)object).getActionCommand(),txtTextEditor.getText()+"");
      else if (object instanceof JToggleButton) {
        // unfortunately a ButtonGroup won't let all buttons be "up" at the same time,
        // so we fake it by doing it ourselves
        detoggleRelationButtons((JToggleButton)object);
        this.pnlGraph.addRelationMode(((JToggleButton)object).getActionCommand(),((JToggleButton)object).isSelected());
      }
    }
  }

  /**
   * sets the Node buttons to enabled or disabled
   * 
   * @param b      the value to use for setEnabled on each button, true enabled, false disabled
   */
  private void setEnableNodeButtons(boolean b) {
    for (int i=0;i<addNodeButtons.size();i++)
      ((JButton)addNodeButtons.get(i)).setEnabled(b);
  }

  /**
   * a java ButtonGroup doesn't allow all buttons to be up, therefore we do this
   * manually by calling this event each time one of the relation buttons is clicked
   * upon, the code will make sure all other buttons are up
   * 
   * @param tb     the JToggleButton that was clicked, all others should be up
   */
  private void detoggleRelationButtons(JToggleButton tb) {
    JToggleButton temp;
    for (int i=0;i<addRelationButtons.size();i++) {
      temp = (JToggleButton)addRelationButtons.get(i);
      if (temp!=tb) temp.setSelected(false);
    }
  }

  /**
   * fired when the model associated with this panel is renamed, comes from
   * the ModelChangeListener, not used for now
   * 
   * @param m      model involved in rename, always the associated model
   */
  public void modelRenamed(Model m) {}

  /**
   * fired when a Constituent is deleted, comes from the ModelChangeListener,
   * not used for now
   * 
   * @param m      model involved in the delete
   * @param p      Constituent deleted
   */
  public void constituentDelete(Model m, Constituent c) {}

  /**
   * fired when a Constituent is inserted, comes from the ModelChangeListener,
   * not used for now
   * 
   * @param m      model involved in the insert
   * @param c      Constituent inserted
   */
  public void constituentInsert(Model m, Constituent c) {}

  /**
   * fired when a Constituent's text is changed, comes from the
   * ModelChangeListener, used to set the text box to reflect the new text
   * 
   * @param m      model involved in the text change
   * @param c      Constituent involved in text change
   */
  public void constituentSetText(Model m, Constituent c) {
    // if it is the selected one then update the text editor
    // unless our text box is the one they are actually typing in
    if (!txtTextEditor.hasFocus()) {
      if (c==mdl.getSelection()) txtTextEditor.setText(c.getText());
    }
  }

  /**
   * fired when a Constituent is experiences an object replace (such as changing
   * from one type of node to another), comes from the ModelChangeListener,
   * not used for now
   * 
   * @param m      model involved in the insert
   * @param c      Constituent inserted
   */
  public void constituentObjectReplace(Model m, Constituent oldCons, Constituent newCons) {}

  /**
   * this method is called via a ModelChangeListener event, it indicates a change
   * in selection to the Constituent c, or null if there is no selection
   * 
   * @param c      the newly selected Constituent of null if none
   */
  public void constituentSelect(Model m, Constituent c) {
    // we don't need to check the model, this GraphViewPanel is only listening to
    // its own model
    if (c==null)
      txtTextEditor.setText("");
    else
      txtTextEditor.setText(c.getText());
    btnDelete.setEnabled(c!=null);
    setEnableNodeButtons(c==null);
  }

  /**
   * this method is called by the application DragDropManager when
   * a drop has occurred on this component, the drop is forwarded to
   * the GraphPanel for this GraphViewPanel
   * 
   * @param x      x coordinate of drop
   * @param y      y coordinate of drop
   * @param o      Object associated with drop
   */
  public void drop(int x, int y, Object o) {
    Point p = pnlGraph.getLocationOnScreen();
    if (y>p.getY()) {
      this.pnlGraph.drop(x,y,o);
    }
  }

}
