/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.io;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import java.io.*;

/* for the xml stuff */
import java.io.IOException;
import org.jdom.*;
import org.jdom.output.XMLOutputter;
import org.jdom.input.SAXBuilder;

import belvedere.model.*;
import belvedere.model.evidence.*;

/**
 * Class used to import belvedere model 3 files
 */
public class Belvedere3Importer {

  //XML constants taken from Belvedere 3
  public static final String BELVEDERE_ROOT = "BelML"; //used by FileTyper
  private static final String PROJECT_ROOT = "Project";
  private static final String MODEL_TYPE = "Evidence_Model";
  private static final String MODEL_ID = "Model_ID";
  private static final String MODEL_NAME = "Model_Name";
  private static final String XML_VERSION = "1.0";
  private static final String SOURCE_NODE_ID = "SourceID";
  private static final String TARGET_NODE_ID = "TargetID";
  private static final String RELATION_ID = "RelationID";
  private static final String CONSTITUENT_ID = "ConstituentID";
  private static final String  POSITION_X = "Postion_X";
  private static final String  POSITION_Y = "Postion_Y";
  private static final String MODELCOMPONENT_URL = "URL";
  private static final String MODELCOMPONENT_MODIFIER = "Modifier";
  private static final String MODELCOMPONENT_CREATOR = "Creator";
  private static final String MODELCOMPONENT_NOTES = "Notes";
  private static final String MODELCOMPONENT_STRENGTH = "Strength";
  private static final String MODELCOMPONENT_JUSTIFICATION = "Justification";
  private static final String RELATION_SOURCE_NAME = "SourceName";
  private static final String RELATION_TARGET_NAME = "TargetName";
  private static final String RELATION_SOURCE_TYPE = "SourceType";
  private static final String RELATION_TARGET_TYPE = "TargetType";
  private static final String RELATION_SOURCE_X = "SOURCE_X";
  private static final String RELATION_SOURCE_Y = "SOURCE_Y";
  private static final String RELATION_TARGET_X = "TARGET_X";
  private static final String RELATION_TARGET_Y = "TARGET_Y";

  //XML constants added
  private static final String DATA_CONSTITUENT = "Data";
  private static final String HYPO_CONSTITUENT = "Hypothesis";
  private static final String UNSPEC_NODE_CONSTITUENT = "UnspecifiedConstituent";
  private static final String CONSISTENCY_CONSTITUENT = "Consistency";
  private static final String INCONSISTENCY_CONSTITUENT = "Inconsistency";
  private static final String UNSPEC_RELATION_CONSTITUENT = "UnspecifiedRelation";
  private static final int NODE_WIDTH = 125;

  /**
   * Attempts to read the given filename which is a Belvedere 3 file and creates
   * and returns a Belvedere 4 project object
   * 
   * @param fileName belvedere 3 file
   * @return the created Belvedere 4 project
   * @exception JDOMException
   * @exception IOException
   */
  public Project readFromXML(String fileName) throws JDOMException, IOException {
    HashMap constituentMap = new HashMap(); //used to locate endpoints of edges
    File belFile = new File(fileName);
    SAXBuilder builder = new SAXBuilder("org.apache.xerces.parsers.SAXParser");
    Document doc = builder.build(belFile);
    Element root = doc.getRootElement();
    // only one project per file supported for now
    Element projElement = (Element)root.getChildren().get(0);
    Project proj = new Project();
    // Belvedere 3 didn't have the Project fields like Belvedere 4
    //  proj.setName(root.getAttributeValue());
    //  proj.setCreator(root.getAttributeValue(Project.CREATOR_KEY));
    //  proj.setCreationDate(root.getAttributeValue(Project.CREATION_DATE_KEY));
    //  proj.setDescription(root.getAttributeValue(Project.DESCR_KEY));
    //  proj.setGUID(root.getAttributeValue(Project.GUID_KEY));
    List modelElements = projElement.getChildren();
    for (Iterator iModels = modelElements.iterator();iModels.hasNext();) {
      Element aModelElement = (Element)iModels.next();
      // only one type in Belvedere 3 for now so check that it matches
      EvidenceModel evModel = new EvidenceModel();
      evModel.setName(aModelElement.getAttributeValue(MODEL_NAME));
      // Belvedere 3 didn't keep track of these fields
      // em.setCreator(aModelElement.getAttributeValue();
      // em.setCreationDate(aModelElement.getAttributeValue(Project.CREATION_DATE_KEY));
      // em.setDescription(aModelElement.getAttributeValue(Project.DESCR_KEY));
      evModel.setGUID(aModelElement.getAttributeValue(MODEL_ID));
      List modelConstituents = aModelElement.getChildren();
      for (Iterator iConstituents = modelConstituents.iterator();iConstituents.hasNext();) {
        Element aConstituentElement = (Element)iConstituents.next();
        if (aConstituentElement.getName().equals(DATA_CONSTITUENT) ||
            aConstituentElement.getName().equals(HYPO_CONSTITUENT) ||
            aConstituentElement.getName().equals(UNSPEC_NODE_CONSTITUENT)) {
          Node n = null;
          if (aConstituentElement.getName().equals(DATA_CONSTITUENT)) {
            n = new DataNode();
          } else if (aConstituentElement.getName().equals(HYPO_CONSTITUENT)) {
            n = new HypothesisNode();
          } else if (aConstituentElement.getName().equals(UNSPEC_NODE_CONSTITUENT)) {
            n = new UnspecifiedNode();
          }
          n.setGUID(aConstituentElement.getAttributeValue(CONSTITUENT_ID));
          n.setX(Integer.parseInt(aConstituentElement.getAttributeValue(POSITION_X)));
          n.setY(Integer.parseInt(aConstituentElement.getAttributeValue(POSITION_Y)));
          n.setURL(aConstituentElement.getAttributeValue(MODELCOMPONENT_URL));
          n.setModifier(aConstituentElement.getAttributeValue(MODELCOMPONENT_MODIFIER));
          n.setCreator(aConstituentElement.getAttributeValue(MODELCOMPONENT_CREATOR));
          n.setNotes(aConstituentElement.getAttributeValue(MODELCOMPONENT_NOTES));
          n.setStrength(Integer.parseInt(aConstituentElement.getAttributeValue(MODELCOMPONENT_STRENGTH)));
          // we don't have Justification
          n.setText(aConstituentElement.getText());
          // they didn't save width in the file
          n.setWidth(NODE_WIDTH);
          evModel.insertVertex(n);
          constituentMap.put(n.getGUID(),n);
        } else if (aConstituentElement.getName().equals(CONSISTENCY_CONSTITUENT) ||
                   aConstituentElement.getName().equals(INCONSISTENCY_CONSTITUENT) ||
                   aConstituentElement.getName().equals(UNSPEC_RELATION_CONSTITUENT)) {
          Relation r = null;
          if (aConstituentElement.getName().equals(CONSISTENCY_CONSTITUENT)) {
            r = new ForRelation();
          } else if (aConstituentElement.getName().equals(INCONSISTENCY_CONSTITUENT)) {
            r = new AgainstRelation();
          } else if (aConstituentElement.getName().equals(UNSPEC_RELATION_CONSTITUENT)) {
            r = new UnspecifiedRelation();
          }
          r.setGUID(aConstituentElement.getAttributeValue(RELATION_ID));
          r.setURL(aConstituentElement.getAttributeValue(MODELCOMPONENT_URL));
          r.setModifier(aConstituentElement.getAttributeValue(MODELCOMPONENT_MODIFIER));
          r.setCreator(aConstituentElement.getAttributeValue(MODELCOMPONENT_CREATOR));
          r.setNotes(aConstituentElement.getAttributeValue(MODELCOMPONENT_NOTES));
          r.setStrength(Integer.parseInt(aConstituentElement.getAttributeValue(MODELCOMPONENT_STRENGTH)));
          //notice that this relies on the endpoints being listed in the XML before the edge that
          //has that endpoint, this is how Belvedere3 works
          Constituent orig = (Constituent)constituentMap.get(aConstituentElement.getAttributeValue(SOURCE_NODE_ID));
          Constituent dest = (Constituent)constituentMap.get(aConstituentElement.getAttributeValue(TARGET_NODE_ID));
          r.setOrigin(orig);
          r.setDestination(dest);
          evModel.insertEdge(r);
          constituentMap.put(r.getGUID(),r);
        }
      }
      proj.addModel(evModel);
    }
    return proj;
  }

}
