/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;

import java.util.HashMap;

import javax.swing.JFrame;

import belvedere.graphview.VisualConstituent;


/**
 * Base class from which all relation and nodes object descend
 */
public abstract class Constituent extends BelvedereObject implements VisualConstituent, Cloneable {

  public static final int STRONGLY_BELIEVE = 5;
  public static final int MODERATELY_BELIEVE = 4;
  public static final int NEUTRALLY_BELIEVE = 3;
  public static final int MODERATELY_UNBELIEVE = 2;
  public static final int STRONGLY_UNBELIEVE = 1;

  // set equal to "" to avoid null situations later
  protected String text = "";
  protected String URL = "";
  protected String notes = "";
  protected String creator = "";
  protected String modifier = "";
  protected int strength = NEUTRALLY_BELIEVE;

  protected Color borderColor = Color.black;
  protected Color regBorderColor = Color.black;
  protected Color selBorderColor = Color.yellow;
  protected Color hovBorderColor = Color.cyan;

  //regBorderStrength comes from strength for now
  //if you do hover and select strokes you will have to store the strengths
  //so they can be restored properly

  //done this way so as to make different strokes for hover/select easy
  //if you want to add it
  protected BasicStroke regBorderStroke = new BasicStroke(strength);
  protected BasicStroke borderStroke = regBorderStroke;

  /**
   * constructor to populate all properties of the constituent
   * 
   * @param text     text for constituent
   * @param URL      URL for constituent
   * @param notes    notes for constituent
   * @param creator  creator for constituent
   * @param modifier modifier for constituent
   * @param strength strength for constituent
   */
  public Constituent(String text, String URL, String notes, String creator,
                     String modifier, int strength) {
    this.text = text;
    this.URL = URL;
    this.notes = notes;
    this.creator = creator;
    this.modifier = modifier;
    // use method call to make sure border stroke is set up
    setStrength(strength);
  }

  /**
   * used to create a HashMap containing the
   * attributes to be saved in the XML representation
   * 
   * @return populated HashMap
   */
  public HashMap getAttributes() {
    HashMap h = new HashMap();
    h.put(Project.TEXT_KEY,this.getText());
    h.put(Project.URL_KEY,this.getURL());
    h.put(Project.NOTES_KEY,this.getNotes());
    h.put(Project.CREATOR_KEY,this.getCreator());
    h.put(Project.MODIFIER_KEY,this.getModifier());
    h.put(Project.STRENGTH_KEY,this.getStrength() + "");
    h.put(Project.GUID_KEY,this.getGUID());
    return h;
  }

  /**
   * Method used to set attributes pulled out of an XML file.
   * Note that unknown attributes that would come from bad
   * XML fall off the end
   * 
   * @param name   name of attribute
   * @param value  value to assign to attribute
   */
  public void setAttribute(String name, String value) {
    if (name.equals(Project.TEXT_KEY)) {
      setText(value);
    } else if (name.equals(Project.URL_KEY)) {
      setURL(value);
    } else if (name.equals(Project.NOTES_KEY)) {
      setNotes(value);
    } else if (name.equals(Project.CREATOR_KEY)) {
      setCreator(value);
    } else if (name.equals(Project.MODIFIER_KEY)) {
      setModifier(value);
    } else if (name.equals(Project.STRENGTH_KEY)) {
      setStrength(Integer.parseInt(value));
    } else if (name.equals(Project.GUID_KEY)) {
      setGUID(value);
    }
  }

  /**
   * sets the text for the constituent
   * 
   * @param text   the text for the constituent
   */
  public void setText(String text) {
    this.text = text;
  }

  /**
   * returns the text of the constituent
   * 
   * @return the text of the constituent
   */
  public String getText() {
    return this.text;
  }

  /**
   * sets the URL for the constituent
   * 
   * @param URL   the URL for the constituent
   */
  public void setURL(String URL) {
    this.URL = URL;
  }

  /**
   * returns the URL of the constituent
   * 
   * @return the URL of the constituent
   */
  public String getURL() {
    return this.URL;
  }

  /**
   * sets the notes for the constituent
   * 
   * @param notes   the notes for the constituent
   */
  public void setNotes(String notes) {
    this.notes = notes;
  }

  /**
   * returns the notes of the constituent
   * 
   * @return the notes of the constituent
   */
  public String getNotes() {
    return this.notes;
  }

  /**
   * sets the creator for the constituent
   * 
   * @param creator   the creator for the constituent
   */    
  public void setCreator(String creator) {
    this.creator = creator;
  }

  /**
   * returns the creator of the constituent
   * 
   * @return the creator of the constituent
   */
  public String getCreator() {
    return this.creator;
  }

  /**
   * sets the modifier for the constituent
   * 
   * @param modifier   the modifier for the constituent
   */    
  public void setModifier(String modifier) {
    this.modifier = modifier;
  }

  /**
   * returns the modifier of the constituent
   * 
   * @return the modifier of the constituent
   */
  public String getModifier() {
    return this.modifier;
  }

  /**
   * sets the strength for the constituent
   * 
   * @param s   the strength for the constituent
   */    
  public void setStrength(int n) {
    this.strength = n;
    // probably should point this to the same stroke or eliminate
    // different stroke states but this needs to be cleaned elsewhere
    // so leave it here for now
    this.regBorderStroke = new BasicStroke(n);
    this.borderStroke = this.regBorderStroke;
  }

  /**
   * returns the strength of the constituent
   * 
   * @return the strength of the constituent
   */
  public int getStrength() {
    return this.strength;
  }

  /**
   * sets the border color for a visual constituent when it is not selected
   * or hovered
   *
   * @param c      the color to set the border to
   */
  public void setBorderColor(Color c) {
    this.regBorderColor = c;
  }

  /**
   * sets the border stroke width for a visual constituent
   * 
   * @param n      the stroke width to set the border to
   */
  public void setBorderStrokeStrength(int n) {
    // NOTE, this also changes the strength of the border as that property
    // is tied to the visual clue of the width of the border!!!
    setStrength(n);
  }

  /**
   * sets the border color for a visual constituent when it is selected
   * 
   * @param c      the color to set the border to
   */
  public void setSelBorderColor(Color c) {
    this.selBorderColor = c;
  }

  /**
   * sets the border color for a visual constituent when it is hovered
   * 
   * @param c      the color to set the border to
   */
  public void setHoverBorderColor(Color c) {
    this.hovBorderColor = c;
  }

  /**
   * this method is called to make the constituent switch it appearance
   *  properties to a selected state
   */
  public void select() {
    this.borderColor = this.selBorderColor;
  }

  /**
   * this method is called to make the constituent switch its appearance
   *  properties to an unselected/not hovered state
   */
  public void deselect() {
    this.borderColor = this.regBorderColor;
    this.borderStroke = this.regBorderStroke;
  }

  /**
   * this method is called to make the constituent switch it appearance
   *  properties to a hovered state
   */
  public void hover() {
    this.borderColor = this.hovBorderColor;
  }

  public String toString() {
    if (this.text.equals(""))
      return "---";
    return this.text;
  }

  /**
   * from this method a constituent will display its property editor
   * 
   * @param parent the parent component that a dialog may use to center over
   */
  public void showPropEditor(JFrame parent, Model mdl) {
    ConstituentPropertyEditor p = new ConstituentPropertyEditor(parent,"Edit Properties",this);
    p.showEditor(mdl);
  }

}
