/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;

import java.util.ArrayList;
import java.util.HashMap;

import javax.swing.JPanel;

import belvedere.graphview.VisualNode;
import belvedere.util.StringParser;

/**
 * the basic node object from which all other nodes should extend
 */
public class Node extends Constituent implements VisualNode {

    //used to hold the location and size of the node
    protected int posX, posY, width, height;

    //used to hold the font properties
    protected Font textFont = new Font("Courier",Font.PLAIN,12);
    protected FontMetrics textFontMetrics;
    protected int textHeight;
    protected int textAscent;
    protected int textInset = 5;  //inset from left border of node

    //used to hold the current properties
    protected Color backColor;
    protected Color textColor;

    //used to hold the regular properties
    protected Color regBackColor = Color.green;
    protected Color regTextColor = Color.black;

    //used to hold the selected properties
    protected Color selBackColor = Color.yellow;
    protected Color selTextColor = Color.black;

    //used to hold the hover properties
    protected Color hovBackColor = Color.cyan;
    protected Color hovTextColor = Color.black;

    //calculated ArrayList of text into rows for visual display
    protected ArrayList displayText = new ArrayList();

    /**
     * Constructor for Node
     */
    public Node() {
      super("","","","","",Constituent.NEUTRALLY_BELIEVE);
      //constituent has selBorderColor yellow, which is fine for relation
      //we change it to black for nodes
      selBorderColor = Color.black;
      hovBorderColor = Color.black;
      //set up current as regular
      borderColor = regBorderColor;
      backColor = regBackColor;
      textColor = regTextColor;
      borderStroke = regBorderStroke;
    }

    /**
     * Constructor for Node
     * 
     * @param text     text for the node
     * @param URL      URL for the node
     * @param notes    notes for the node
     * @param creator  creator of the node
     * @param modifier modifier of the node
     * @param strength strength of the node, indicated in line width
     */
    public Node(String text, String URL, String notes, String creator,
                       String modifier, int strength) {
      super(text,URL,notes,creator,modifier,strength);
      //constituent has selBorderColor yellow, which is fine for relation
      //we change it to black for nodes
      selBorderColor = Color.black;
      hovBorderColor = Color.black;
      //set up current as regular
      borderColor = regBorderColor;
      backColor = regBackColor;
      textColor = regTextColor;
      borderStroke = regBorderStroke;
    }
    /**
     * Constructor for Node
     * 
     * @param text     text for the node
     * @param URL      URL for the node
     * @param notes    notes for the node
     * @param creator  creator of the node
     * @param modifier modifier of the node
     * @param strength strength of the node, indicated in line width
     * @param x        x coordinate for node
     * @param y        y coordinate for node
     * @param w        width for node
     */
    public Node(String text, String URL, String notes, String creator,
                       String modifier, int strength, int x, int y, int width) {
        this(text, URL, notes, creator, modifier, strength);
        posX = x;
        posY = y;
        this.width = width;
        setDisplayText();
    }


    /**
     * Returns a new <code>Node</code> object which is a copy of this instance.
     *
     * @return clone  a copy <code>Object</code> instance of this <code>Node</code>.
     */
    public Object clone() {
      return new Node(this.text, this.URL, this.notes, 
                      this.creator, this.modifier, this.strength, 
                      this.posX, this.posY, this.width);
    }

    /**
     * used to create a HashMap containing the
     * attributes to be saved in the XML representation
     * overrides the constituents getAttributes to add id, xccor, ycoor, width
     * 
     * @return populated HashMap
     */
    public HashMap getAttributes() {
      HashMap h = super.getAttributes();
      h.put(Project.XCOOR_KEY,this.getX() + "");
      h.put(Project.YCOOR_KEY,this.getY() + "");
      h.put(Project.WIDTH_KEY,this.getWidth() + "");
      return h;
    }

    /**
     * Method used to set attributes pulled out of an XML file.
     * Note that unknown attributes that would come from bad
     * XML fall off the end
     * 
     * @param name   name of attribute
     * @param value  value to assign to attribute
     */
    public void setAttribute(String name, String value) {
      super.setAttribute(name,value);
      if (name.equals(Project.XCOOR_KEY)) {
        this.setX(Integer.parseInt(value));
      } else if (name.equals(Project.YCOOR_KEY)) {
        this.setY(Integer.parseInt(value));
      } else if (name.equals(Project.WIDTH_KEY)) {
        this.setWidth(Integer.parseInt(value));
      }
    }

    /**
     * sets the x coordinate for this node
     * 
     * @param x      new x coordinate
     */
    public void setX(int x) {
        posX = x;
    }

    /**
     * sets the y coordinate for this node
     * 
     * @param y      the new y coordinate
     */
    public void setY(int y) {
        posY = y;
    }

    /**
     * returns the x coordinate for this node
     * 
     * @return x coordinate
     */
    public int getX() {
        return posX;
    }

    /**
     * returns the y coordinate for this node
     * 
     * @return y coordinate
     */
    public int getY() {
        return posY;
    }

    /**
     * returns the horizontal center of this node
     * 
     * @return horizontal center
     */
    public int getCenterX() {
        return posX+(width/2);
    }

    /**
     * returns the vertical center of this node
     * 
     * @return vertical center
     */
    public int getCenterY() {
        return posY+(height/2);
    }

    /**
     * sets the width of this node
     * 
     * @param width  new width for node
     */
    public void setWidth(int width) {
        this.width = width;
        setDisplayText();
    }

    /**
     * sets the height for this node
     * 
     * @param height new height
     */
    public void setHeight(int height) {
        this.height = height;
    }

    /**
     * returns the width for this node
     * 
     * @return width of node
     */
    public int getWidth() {
        return width;
    }

    /**
     * returns the height of this node
     * 
     * @return node height
     */
    public int getHeight() {
        return height;
    }

    /**
     * sets the regular back color for the node
     * 
     * @param c      new back color
     */
    public void setBackColor(Color c) {
        regBackColor = c;
    }

    /**
     * sets the selected back color for the node
     * 
     * @param c      new selected back color
     */
    public void setSelBackColor(Color c) {
        selBackColor = c;
    }

    /**
     * sets the text color for the node
     * 
     * @param c      new text color
     */
    public void setTextColor(Color c) {
        textColor = c;
    }

    /**
     * sets the selected text color for the node
     * 
     * @param c      new selected text color
     */
    public void setSelTextColor(Color c) {
        selTextColor = c;
    }

    /**
     * sets the back color for a visual node when it is hovered
     * 
     * @param c      the color to set the back to
     */    
    public void setHoverBackColor(Color c) {
        hovBackColor = c;
    }


    /**
     * puts the node in the select state, changing its current properties
     * to match the values for selected properties
     */
    public void select() {
        borderColor = selBorderColor;
        backColor = selBackColor;
        textColor = selTextColor;
    }

    /**
     * takes a node out of the select state by returning its current properties
     * to the regular properties
     */
    public void deselect() {
        borderColor = regBorderColor;
        backColor = regBackColor;
        textColor = regTextColor;
    }

    /**
     * this method is called to make the constituent switch it appearance
     *  properties to a hovered state
     */
    public void hover() {
        borderColor = hovBorderColor;
        backColor = hovBackColor;
        textColor = hovTextColor;
    }


    /**
     * sets the text for a node
     * 
     * @param text   new text for a node
     */
    public void setText(String text) {
        this.text = text;
        //now need to set up the displayText whenever the text, font,
        //or dimensions of the bounding region change so that it doesn't have
        //to be done on every paint();
        setDisplayText();
    }

    /**
     * used to set the display text for the node.  This is where the word wrapping is
     * calculated.
     */
    private void setDisplayText() {
        //this method will set up the displayText whenever the text, font,
        //or dimensions of the bounding region change so that it doesn't have
        //to be done on every paint();

        //a temporary jpanel is necessary to get the font metrics
        JPanel temp = new JPanel();
        textFontMetrics = temp.getFontMetrics(textFont);
        textHeight = textFontMetrics.getHeight();
        textAscent = textFontMetrics.getAscent();

        //NOTE max rows hard coded for now, should be passed in the constructor
        //and height should not be passed in
        
        displayText = StringParser.parseFromFront(text,textFontMetrics,width - textInset*2,5);

        height = displayText.size()*textFontMetrics.getHeight() + 2*textInset;
    }

    /**
     * determining whether the given point is within the Constituents rect
     * 
     * @param x      x coordinate of point to test
     * @param y      y coordinate of point to test
     * @return true if x,y is inside the node, false otherwise
     */
    public boolean withinBounds(int x, int y) {
    return (x>posX && x<posX+width) && (y>posY && y<posY+height);
    }

    /**
     * used to paint the visual node on the given graphics
     * 
     * @param g      graphics to paint upon
     */
    public void paint(Graphics g) {
        Graphics2D g2D = (Graphics2D)g;
        g2D.setStroke(borderStroke);
        g2D.setColor(backColor);
        g2D.fillRect(posX,posY,width,height); //draw the background
        g2D.setColor(borderColor);
        g2D.drawRect(posX,posY,width,height); //draw the border
        g2D.setColor(textColor);
        g2D.setFont(textFont);
        int textPosX = posX + textInset;
        int textPosY = posY + textAscent + textInset;
        for (int i=0;i<displayText.size();i++)
            g2D.drawString((String)displayText.get(i),textPosX,textPosY + (i*textHeight));
    }
}
