/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Iterator;
import java.util.Enumeration;

/* for the xml stuff */
import java.io.IOException;
import org.jdom.*;
import org.jdom.output.XMLOutputter;
import org.jdom.input.SAXBuilder;

import java.io.File;
import java.io.FileOutputStream;

import belvedere.util.ObjectFactory;

/**
 * A Project represents a collection of Models.
 */
public class Project extends BelvedereObject {

  //attribrute names used in XML of a Belvedere Project
  public static final String PROG_NAME_KEY = "belvedere";
  public static final String PROG_VER_KEY = "version";
  public static final String NAME_KEY = "name";
  public static final String CREATOR_KEY = "creator";
  public static final String CREATION_DATE_KEY = "creation_date";
  public static final String DESCR_KEY = "description";
  public static final String TEXT_KEY = "text";
  public static final String URL_KEY = "url";
  public static final String NOTES_KEY = "notes";
  public static final String MODIFIER_KEY = "modifier";
  public static final String STRENGTH_KEY = "strength";
  public static final String XCOOR_KEY = "xcoor";
  public static final String YCOOR_KEY = "ycoor";
  public static final String WIDTH_KEY = "width";
  public static final String ORIGIN_GUID_KEY = "origin_guid";
  public static final String DEST_GUID_KEY = "destination_guid";
  public static final String NODES_KEY = "nodes";
  public static final String RELATIONS_KEY = "relations";
  public static final String GUID_KEY = "guid";

  private ArrayList projectChangeListeners = new ArrayList();

  private boolean dirty = false;

  // they put "class " when returning class name, this is used to skip it
  private static final int CLASS_SKIP = 6;

  private static final String XML_INDENT = "  ";

  private static final String VERSION = "4.0";

  private ArrayList models; //collection of all models in this project.
  private String name; //String identifier of this project.
  private String creator;
  private String creationDate;
  private String description;

  /**
   * Main constructor.  Creates a Project with the given name.
   *
   * @param name  Name of Project being created.
   */
  public Project(String name) {
     this.models = new ArrayList();
     this.name = name;
     this.creator = "";
     this.creationDate = "";
     this.description = "";
  }

  /**
   * Default constructor, creates a project without a name.
   */
  public Project() {
    this("");
  }

  /**
   * saves the project in xml format to the file filename
   * 
   * @param filename filename for xml file
   * @exception IOException
   */
  public void save(String filename) throws IOException {
    Element projectElement = new Element(Project.PROG_NAME_KEY);
    projectElement.setAttribute(Project.PROG_VER_KEY,VERSION);
    projectElement.setAttribute(Project.NAME_KEY,this.getName());
    projectElement.setAttribute(Project.CREATOR_KEY,this.getCreator());
    projectElement.setAttribute(Project.CREATION_DATE_KEY,this.getCreationDate());
    projectElement.setAttribute(Project.DESCR_KEY,this.getDescription());
    projectElement.setAttribute(Project.GUID_KEY,this.getGUID());
    for (Iterator iModels=models.iterator();iModels.hasNext();) {
      Model aModel = (Model)iModels.next();
      Element modelElement = new Element(aModel.getClass().toString().substring(CLASS_SKIP));
      modelElement.setAttribute(Project.NAME_KEY,aModel.getName());
      modelElement.setAttribute(Project.CREATOR_KEY,aModel.getCreator());
      modelElement.setAttribute(Project.CREATION_DATE_KEY,aModel.getCreationDate());
      modelElement.setAttribute(Project.DESCR_KEY,aModel.getDescription());
      modelElement.setAttribute(Project.GUID_KEY,aModel.getGUID());
      Element nodeElements = new Element(Project.NODES_KEY);
      for (Iterator iter = aModel.vertices();iter.hasNext();) {
        Node n = (Node)iter.next();
        Element nodeElement = new Element(n.getClass().toString().substring(CLASS_SKIP));
        for (Iterator iAttributes = n.getAttributes().entrySet().iterator();iAttributes.hasNext();) {
          Entry e = (Entry)iAttributes.next();
          nodeElement.setAttribute(e.getKey().toString(),e.getValue().toString());
        }
        nodeElements.addContent(nodeElement);
      }
      modelElement.addContent(nodeElements);
      Element relationElements = new Element(Project.RELATIONS_KEY);
      for (Iterator iter = aModel.edges();iter.hasNext();) {
        Relation r = (Relation)iter.next();
        Element relationElement = new Element(r.getClass().toString().substring(CLASS_SKIP));
        for (Iterator iAttributes = r.getAttributes().entrySet().iterator();iAttributes.hasNext();) {
          Entry e = (Entry)iAttributes.next();
          relationElement.setAttribute(e.getKey().toString(),e.getValue().toString());
        }
        relationElements.addContent(relationElement);
      }
      modelElement.addContent(relationElements);
      projectElement.addContent(modelElement);
    }
    Document doc = new Document(projectElement);
    XMLOutputter out = new XMLOutputter(XML_INDENT,true);
    out.output(doc, new FileOutputStream(filename));
  }

  /**
   * reads a belvedere project from an xml file
   * 
   * @param filename file to read
   * @exception IOException
   * @exception JDOMException
   */
  public void read(String filename) throws IOException, JDOMException {
    HashMap constituentMap = new HashMap();
    File belFile = new File(filename);
    SAXBuilder builder = new SAXBuilder("org.apache.xerces.parsers.SAXParser");
    Document doc = builder.build(belFile);
    Element root = doc.getRootElement();
    if (!root.getName().equals(Project.PROG_NAME_KEY))
      throw new JDOMException("File does not appear to be a valid Belvedere file");
    this.setName(root.getAttributeValue(Project.NAME_KEY));
    this.setCreator(root.getAttributeValue(Project.CREATOR_KEY));
    this.setCreationDate(root.getAttributeValue(Project.CREATION_DATE_KEY));
    this.setDescription(root.getAttributeValue(Project.DESCR_KEY));
    this.setGUID(root.getAttributeValue(Project.GUID_KEY));
    List modelElements = root.getChildren();
    for (Iterator iModels = modelElements.iterator();iModels.hasNext();) {
      Element aModelElement = (Element)iModels.next();
      Model model = (Model)ObjectFactory.createObject(aModelElement.getName());
      if (model==null)
        throw new JDOMException("File is invalid");
      model.setName(aModelElement.getAttributeValue(Project.NAME_KEY));
      model.setCreator(aModelElement.getAttributeValue(Project.CREATOR_KEY));
      model.setCreationDate(aModelElement.getAttributeValue(Project.CREATION_DATE_KEY));
      model.setDescription(aModelElement.getAttributeValue(Project.DESCR_KEY));
      model.setGUID(aModelElement.getAttributeValue(Project.GUID_KEY));
      Element nodeHolder = aModelElement.getChild(Project.NODES_KEY);
      List nodeElements = nodeHolder.getChildren();
      for (Iterator iNodes = nodeElements.iterator();iNodes.hasNext();) {
        Element aNodeElement = (Element)iNodes.next();
        Node node = (Node)ObjectFactory.createObject(aNodeElement.getName());
        if (node==null)
          throw new JDOMException("File is invalid");
        /* loop through the attributes passing them to setAttribute */
        for (Iterator iAttributes = aNodeElement.getAttributes().iterator();iAttributes.hasNext();) {
          Attribute a = (Attribute)iAttributes.next();
          node.setAttribute(a.getName(),a.getValue());
        }
        model.insertVertex(node);
        constituentMap.put(node.getGUID(),node);
      }
      Element relationHolder = aModelElement.getChild(Project.RELATIONS_KEY);
      List relationElements = relationHolder.getChildren();
      for (Iterator iRelations = relationElements.iterator();iRelations.hasNext();) {
        Element aRelationElement = (Element)iRelations.next();
        Relation relation = (Relation)ObjectFactory.createObject(aRelationElement.getName());
        if (relation==null)
          throw new JDOMException("File is invalid");
        /* loop through the attributes passing them to setAttribute */
        for (Iterator iAttributes = aRelationElement.getAttributes().iterator();iAttributes.hasNext();) {
          Attribute a = (Attribute)iAttributes.next();
          relation.setAttribute(a.getName(),a.getValue());
        }
        Constituent orig = (Constituent)constituentMap.get(aRelationElement.getAttributeValue(Project.ORIGIN_GUID_KEY));
        Constituent dest = (Constituent)constituentMap.get(aRelationElement.getAttributeValue(Project.DEST_GUID_KEY));
        relation.setOrigin(orig);
        relation.setDestination(dest);
        model.insertEdge(relation);
        constituentMap.put(relation.getGUID(),relation);
      }
      this.addModel(model);
    }
  }

  /**
   * adds an object implementing the ProjectChangeListener interface to the
   * list of listeners notified of project change events
   * 
   * @param pc     the ProjectChangeListener to add to the list
   */
  public void addProjectChangeListener(ProjectChangeListener pc) {
    projectChangeListeners.add(pc);
  }

  /**
   * Returns the name of this project.
   *
   * @return name  name of this project
   */
  public String getName() {
    return this.name;
  }

  /**
   * Sets the name of this project.
   *
   * @param name   Text String naming this project
   */
  public void setName(String name) {
    this.name = name;
  }

  /**
   * Returns the Creator of this project.
   *
   * @return Creator  Creator of this project
   */
  public String getCreator() {
    return this.creator;
  }

  /**
   * Sets the creator of this project.
   *
   * @param creator   Text String naming this project
   */
  public void setCreator(String creator) {
    this.creator = creator;
  }

  /**
   * Returns the creationDate of this project.
   *
   * @return creationDate  creationDate of this project
   */
  public String getCreationDate() {
    return this.creationDate;
  }

  /**
   * Sets the creationDate of this project.
   *
   * @param creationDate   Text String for creationDate of project
   */
  public void setCreationDate(String creationDate) {
    this.creationDate = creationDate;
  }

  /**
   * Returns the description of this project.
   *
   * @return description  description of this project
   */
  public String getDescription() {
    return this.description;
  }

  /**
   * Sets the description of this project.
   *
   * @param description   Text String naming this project
   */
  public void setDescription(String description) {
    this.description = description;
  }

  /**
   * Adds a <code>Model</code> to this Project.
   *
   * @param newModel   Model to be added to Project.
   */
  public void addModel(Model newModel) {
    this.models.add(newModel);
    this.fireModelAdded(newModel);
  }

  
  /**
   * Removes a <code>Model</code> from this Project.
   *
   * @param oldModel    Model to be removed.
   * @return boolean   true if successfully removed the model. 
   */
  public boolean removeModel(Model oldModel) { 
    this.fireModelRemoved(oldModel,models.indexOf(oldModel));
    return this.models.remove(oldModel);
  }


  /**
   * Removes a <code>Model</code> from this Project.
   *
   * @param index    integer index of <code>Model</code> to be removed.
   * @return model   returns the <code>Model</code> at 
   *                 index if successfully removed, null otherwise. 
   */
  public Model removeModel(int index) {
    this.fireModelRemoved((Model)models.get(index),index);
    return (Model)this.models.remove(index);
  }


  /**
   * Returns a <code>java.util.Collection<code> containing all <code>Models</code>
   * in this project.
   *
   * @return models   Collection of all <code>Model</code> Objects in this Project.
   */
  public Collection getModels() {
    return this.models;
  }


  /**
   * Attempts to retrieve the model with the requested name from this project.
   *
   * @param name  Name of model which is being retrieved.
   * @return model  Model object which was successfully retrieved.
   * @exception ModelNotFoundException thrown if model is not retrieved.
   */
  public Model getModelByName(String name) throws ModelNotFoundException {
    Model temp;
    for (int i = 0; i < models.size(); i++) {
       if ((temp = (Model)models.get(i)).getName().equals(name)) 
          return temp;
    }
    throw new ModelNotFoundException("The requested model could not be " +
                                     "found in this project");
  }

  /**
   * Determines whether or not the project is in a dirty state.  Used as an
   * indicator for unsaved changes.
   * 
   * @return true if dirty, false otherwise
   */
  public boolean isDirty() {
      //check the project dirty first
      if (this.dirty) return true;
      //now check the models for a dirty one
      for (int i=0;i<models.size();i++)
          if (((Model)models.get(i)).isDirty())              
              return true;
      // none of the models were dirty, so return false
      return false;
  }

  /**
   * Sets the dirty flag for a Project
   * 
   * @param d      the value to set the dirty flag to
   */
  public void setDirty(boolean d) {
      //set dirty for the project
      this.dirty = d;
      //if d was true, the above statement is enough to set the project dirty,
      //because isDirty() checks the models, for false we must also set models
      if (!d) {
          for (int i=0;i<models.size();i++)
              ((Model)models.get(i)).setDirty(d);
      }
  }

  /**
   * used to notify the ProjectChangeListeners of a model removed event
   * 
   */
  protected void fireModelRemoved(Model m, int index) {
    for (int i=0;i<projectChangeListeners.size();i++)
      ((ProjectChangeListener)projectChangeListeners.get(i)).modelRemoved(this,m,index);
  }

  /**
   * used to notify the ProjectChangeListeners of a model added event
   * 
   */
  protected void fireModelAdded(Model m) {
    for (int i=0;i<projectChangeListeners.size();i++)
      ((ProjectChangeListener)projectChangeListeners.get(i)).modelAdded(this,m);
  }

}
