/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model.evidence;

import java.awt.Shape;
import java.awt.Rectangle;
import java.awt.Point;
import java.awt.Dimension;
import java.awt.geom.Rectangle2D;
import java.awt.geom.PathIterator;
import java.awt.geom.AffineTransform;
import java.awt.geom.IllegalPathStateException;
import sun.awt.geom.Curve;



public final class Cloud extends Rectangle {


  private static final int DEFAULT_WIDTH = 160;
  private static final int DEFAULT_HEIGHT = 32;
  private static final int ARCH_MASK = 0xFFFFFFF8; //verifies a multiple of 8.  
  private static final int ARCH_LENGTH = 8;  // length of arch, set mask also
  private static final int ARCH_POINT = 8;  // distance used for setting arch.


  /**** the following is taken from java.awt.geom.General Path  ****/

  // For code simplicity, copy these constants to our namespace
  // and cast them to byte constants for easy storage.
  private static final byte SEG_MOVETO  = (byte) PathIterator.SEG_MOVETO;
  private static final byte SEG_LINETO  = (byte) PathIterator.SEG_LINETO;
  private static final byte SEG_QUADTO  = (byte) PathIterator.SEG_QUADTO;
  
  private static final int INIT_SIZE = 20;
  private static final int EXPAND_MAX = 500;

  byte[] pointTypes;
  float[] pointCoords;
  int numTypes;
  int numCoords;
  int windingRule;
 /****  end java.awt.geom.GeneralPath excerpt ****/

  /**
   *
   *
   */
  public Cloud() {
    this(0,0, DEFAULT_WIDTH, DEFAULT_HEIGHT); 
  }

  /**
   *
   *
   */
  public Cloud(Rectangle r) {
    this(r.x, r. y, r.width, r.height);
  }

  /**
   *
   *
   */
  public Cloud(int x, int y, int width, int height) {
    this.x = x;
    this.y = y;
    this.width = width & ARCH_MASK;
    this.height = height & ARCH_MASK; 
    this.windingRule = PathIterator.WIND_NON_ZERO;
    this.pointTypes = new byte[INIT_SIZE];
    this.pointCoords = new float[INIT_SIZE * 2];

  }

  /**
   *
   *
   */
  public Cloud(int width, int height) {
    this(0, 0, width, height);
  }

  /**
   *
   *
   */
  public Cloud(Point p) {
    this(p.x, p.y, DEFAULT_WIDTH, DEFAULT_HEIGHT);
  }

  /**
   *
   *
   */
  public Cloud(Point p, Dimension d) {
    this(p.x, p.y, d.width, d.height);
  }

  /**
   *
   *
   */
  public Cloud(Dimension d) {
    this(0, 0, d.width, d.height);
  }

  /**
   * Sets the size of this <code>Cloud</code> to match the 
   * specified <code>Dimension</code>.
   *
   * @param d the new size for the <code>Cloud</code> object
   * @see       java.awt.Component#setSize(java.awt.Dimension)
   */
  public void setSize(Dimension d) {
    this.setSize(d.width, d.height);
  } 

  /**
   * Sets the size of this <code>Cloud</code> to the specified 
   * width and height.
   * 
   * @param width the new width for this <code>Cloud</code>
   * @param height the new height for this <code>Cloud</code>
   * @see       java.awt.Component#setSize(int, int)
   */
  public void setSize(int width, int height) {
    super.resize(width & this.ARCH_MASK, height & this.ARCH_MASK);
  } 

  /**
   * Sets the bounding <code>Rectangle</code> of this <code>Cloud</code> 
   * to match the specified <code>Rectangle</code>.
   * 
   * @param r   the specified <code>Rectangle</code>
   * @see       java.awt.Component#setBounds(java.awt.Rectangle)
   */
  public void setBounds(Rectangle r) {
    this.setBounds(r.x, r.y, r.width & this.ARCH_MASK, 
                   r.height & this.ARCH_MASK);
  }

  /**
   * Sets the bounding <code>Rectangle</code> of this 
   * <code>Cloud</code> to the specified 
   * <code>x</code>, <code>y</code>, <code>width</code>, 
   * and <code>height</code>.
   * 
   * @param x,&nbsp;y the new coordinates for the top-left
   *                    corner of this <code>Cloud</code>
   * @param width the new width for this <code>Cloud</code>
   * @param height the new height for this <code>Cloud</code>
   * @see       java.awt.Component#setBounds(int, int, int, int)
   */
  public void setBounds(int x, int y, int width, int height) {
    reshape(x, y, width & this.ARCH_MASK, height & this.ARCH_MASK);
  } 


  /**
   *
   *
   */
  public void setRect(double x, double y, double width, double height) {
    super.setRect(x, y, (int)width & this.ARCH_MASK, 
                  (int)height & this.ARCH_MASK);  
  }


  /**
   *
   *
   */
  public void setWidth(int width) {
    this.width = width & this.ARCH_MASK;
  }


  /**
   *
   *
   */
  public void setHeight(int height) {
    this.height = height & this.ARCH_MASK;
  }


  /**
   * Establishes a sequence of arcs as a border of the rectangular shape.
   *
   */
  public void setBorder() {
    int x1, x2, y1, y2;
    x1 = this.x;
    y1 = this.y;
    x2 = x1 + this.width;
    y2 = y1 + this.height;

    // establish top of cloud
    for (; x1 < x2; x1+=this.ARCH_LENGTH) {
      this.moveTo(x1, y1);
      this.quadTo(x1+this.ARCH_LENGTH/2, y1-this.ARCH_POINT, x1+this.ARCH_LENGTH, y1);
    }
    // establish right side of cloud
    for (; y1 < y2; y1+=this.ARCH_LENGTH) {
      this.moveTo(x1, y1);
      this.quadTo(x1+this.ARCH_POINT, y1+this.ARCH_LENGTH/2, x1, y1+this.ARCH_LENGTH);
    }     
    // establish bottom of cloud
    for (; x1 > this.x; x1-=this.ARCH_LENGTH) {
      this.moveTo(x1, y1);
      this.quadTo(x1-this.ARCH_LENGTH/2, y1+this.ARCH_POINT, x1-this.ARCH_LENGTH, y1);
    }
    // establish left side of cloud
    for (; y1 > this.y; y1-=this.ARCH_LENGTH) {
      this.moveTo(x1, y1);
      this.quadTo(x1-this.ARCH_POINT, y1-this.ARCH_LENGTH/2, x1, y1-this.ARCH_LENGTH);
    }
  }

  
  /**
   * Adds the contained rectangular shape to the border, so this cloud consists 
   * of the complete internal area of the border.
   *
   */
  public void fillShape() {
     this.moveTo(this.x, this.y);
     this.lineTo(this.x + this.width, this.y);
     this.lineTo(this.x, this.y + this.height);
     this.moveTo(this.x + this.width, this.y + this.height);
     this.lineTo(this.x + this.width, this.y);
     this.lineTo(this.x, this.y + this.height);
  }
  
  /***  the following is taken from java.awt.geom.GeneralPath ***/

  private void needRoom(int newTypes, int newCoords, boolean needMove) {
    if (needMove && numTypes == 0) {
      throw new IllegalPathStateException("missing initial moveto "+
                                          "in path definition");
    }
    int size = pointCoords.length;
    if (numCoords + newCoords > size) {
      int grow = size;
      if (grow > EXPAND_MAX * 2) {
        grow = EXPAND_MAX * 2;
      }
      if (grow < newCoords) {
        grow = newCoords;
      }
      float[] arr = new float[size + grow];
      System.arraycopy(pointCoords, 0, arr, 0, numCoords);
      pointCoords = arr;
    }
    size = pointTypes.length;
    if (numTypes + newTypes > size) {
      int grow = size;
      if (grow > EXPAND_MAX) {
        grow = EXPAND_MAX;
      }
      if (grow < newTypes) {
        grow = newTypes;
      }
      byte[] arr = new byte[size + grow];
      System.arraycopy(pointTypes, 0, arr, 0, numTypes);
      pointTypes = arr;
    }
  }


  /**
   * Adds a point to the path by moving to the specified
   * coordinates.
   * @param x,&nbsp;y the specified coordinates
   */
  public synchronized void moveTo(float x, float y) {
    if (numTypes > 0 && pointTypes[numTypes - 1] == SEG_MOVETO) {
      pointCoords[numCoords - 2] = x;
      pointCoords[numCoords - 1] = y;
    } else {
      needRoom(1, 2, false);
      pointTypes[numTypes++] = SEG_MOVETO;
      pointCoords[numCoords++] = x;
      pointCoords[numCoords++] = y;
    }
  }


  /**
   * Adds a point to the path by drawing a straight line from the
   * current coordinates to the new specified coordinates.
   * @param x,&nbsp;y the specified coordinates
   */
  public synchronized void lineTo(float x, float y) {
    needRoom(1, 2, true);
    pointTypes[numTypes++] = SEG_LINETO;
    pointCoords[numCoords++] = x;
    pointCoords[numCoords++] = y;
  }


  /**
   * Adds a curved segment, defined by two new points, to the path by
   * drawing a Quadratic curve that intersects both the current
   * coordinates and the coordinates (x2,&nbsp;y2), using the 
   * specified point (x1,&nbsp;y1) as a quadratic parametric control
   * point.
   * @param x1,&nbsp;y1 the coordinates of the first quadratic control
   *    point
   * @param x2,&nbsp;y2 the coordinates of the final endpoint
   */
  public synchronized void quadTo(float x1, float y1, float x2, float y2) {
    needRoom(1, 4, true);
    pointTypes[numTypes++] = SEG_QUADTO;
    pointCoords[numCoords++] = x1;
    pointCoords[numCoords++] = y1;
    pointCoords[numCoords++] = x2;
    pointCoords[numCoords++] = y2;
  }


  /**
   * Tests if the specified coordinates are inside the boundary of 
   * this <code>Shape</code>.
   * @param x,&nbsp;y the specified coordinates
   * @return <code>true</code> if the specified coordinates are inside this 
   * <code>Shape</code>; <code>false</code> otherwise
   */
  public boolean contains(double x, double y) {
    if (numTypes < 2) {
      return false;
    }
    int cross = Curve.crossingsForPath(getPathIterator(null), x, y);
    if (windingRule == PathIterator.WIND_NON_ZERO) {
      return(cross != 0);
    } else {
      return((cross & 1) != 0);
    }
  }

  /**
   * Sets the winding rule for this path to the specified value.
   * @param rule an integer representing the specified 
   * winding rule
   * @exception <code>IllegalArgumentException</code> if 
   *    <code>rule</code> is not either 
   *    <code>WIND_EVEN_ODD</code> or
   *    <code>WIND_NON_ZERO</code>
   * @see #WIND_EVEN_ODD  
   * @see #WIND_NON_ZERO
   */
  public void setWindingRule(int rule) {
    if (rule != PathIterator.WIND_EVEN_ODD && rule != PathIterator.WIND_NON_ZERO) {
      throw new IllegalArgumentException("winding rule must be "+
                                         "WIND_EVEN_ODD or "+
                                         "WIND_NON_ZERO");
    }
    windingRule = rule;
  }

  /**
   * Returns the fill style winding rule.
   * @return an integer representing the current winding rule.
   * @see #WIND_EVEN_ODD  
   * @see #WIND_NON_ZERO
   */
  public synchronized int getWindingRule() {
    return windingRule;
  }

  /**
 * Resets the path to empty.  The append position is set back to the
 * beginning of the path and all coordinates and point types are
 * forgotten.
 */
  public synchronized void reset() {
    numTypes = numCoords = 0;
  }
    /****  end java.awt.geom.GeneralPath excerpt ****/
  
  
  /**
   * Returns an iteration object that defines the boundary of this
   * <code>Cloud</code>.
   * The iterator for this class is multi-threaded safe, which means
   * that this <code>Cloud</code> class guarantees that
   * modifications to the geometry of this <code>Cloud</code>
   * object do not affect any iterations of that geometry that
   * are already in process.
   *
   * @param at an optional <code>AffineTransform</code> to be applied to
   * the coordinates as they are returned in the iteration, or
   * <code>null</code> if untransformed coordinates are desired
   * @return    the <code>PathIterator</code> object that returns the
   *          geometry of the outline of this
   *          <code>Rectangle2D</code>, one segment at a time.
   * @since 1.2
   */
  public PathIterator getPathIterator(AffineTransform at) {
    return new CloudPathIterator(this, at);
  }

  /**
   * Returns an iteration object that defines the boundary of the
   * flattened <code>Rectangle2D</code>.  Since rectangles are already
   * flat, the <code>flatness</code> parameter is ignored.
   * The iterator for this class is multi-threaded safe, which means
   * that this <code>Rectangle2D</code> class guarantees that
   * modifications to the geometry of this <code>Rectangle2D</code>
   * object do not affect any iterations of that geometry that
   * are already in process.
   * @param at an optional <code>AffineTransform</code> to be applied to
   * the coordinates as they are returned in the iteration, or
   * <code>null</code> if untransformed coordinates are desired
   * @param flatness the maximum distance that the line segments used to
   * approximate the curved segments are allowed to deviate from any
   * point on the original curve.  Since rectangles are already flat,
   * the <code>flatness</code> parameter is ignored.
   * @return    the <code>PathIterator</code> object that returns the
   *          geometry of the outline of this
   *          <code>Rectangle2D</code>, one segment at a time.
   * @since 1.2
   */
  public PathIterator getPathIterator(AffineTransform at, double flatness) {
    return new CloudPathIterator(this, at);
  }


}



