/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model.evidence;

import java.awt.geom.*;

/**
 * This class is essentially the same as  
 * <code>java.awt.geom.GeneralPathIterator<code> except it has been modified
 * to work with a specific shape: <code>belvedere.model.evidence.Cloud</code>.
 */                                                                  
public class CloudPathIterator implements PathIterator {

  private int typeIdx = 0;
  private int pointIdx = 0;
  private Cloud cloudPath;
  private AffineTransform affine;
  private static final int curvesize[] = {2, 2, 4, 6, 0};


  public CloudPathIterator(Cloud cloudPath) {
    this(cloudPath, null);
  }

  public CloudPathIterator(Cloud cloudPath, AffineTransform at) {
    this.cloudPath = cloudPath;
    this.affine = at;
  }

  /**
  * Return the winding rule for determining the interior of the
  * cloud.
  * @see PathIterator#WIND_EVEN_ODD
  * @see PathIterator#WIND_NON_ZERO
  */
  public int getWindingRule() {
    return cloudPath.getWindingRule();
  }

  /**
   * Tests if there are more points to read.
   * @return true if there are more points to read
   */
  public boolean isDone() {
    return(typeIdx >= cloudPath.numTypes);
  }

  /**
   * Moves the iterator to the next segment of the path forwards
   * along the primary direction of traversal as long as there are
   * more points in that direction.
   */
  public void next() {
    int type = cloudPath.pointTypes[typeIdx++];
    pointIdx += curvesize[type];
  }

  /**
   * Returns the coordinates and type of the current path segment in
   * the iteration.
   * The return value is the path segment type:
   * SEG_MOVETO, SEG_LINETO, SEG_QUADTO, SEG_CUBICTO, or SEG_CLOSE.
   * A float array of length 6 must be passed in and may be used to
   * store the coordinates of the point(s).
   * Each point is stored as a pair of float x,y coordinates.
   * SEG_MOVETO and SEG_LINETO types will return one point,
   * SEG_QUADTO will return two points,
   * SEG_CUBICTO will return 3 points
   * and SEG_CLOSE will not return any points.
   * @see PathIterator#SEG_MOVETO
   * @see PathIterator#SEG_LINETO
   * @see PathIterator#SEG_QUADTO
   * @see PathIterator#SEG_CUBICTO
   * @see PathIterator#SEG_CLOSE
   */
  public int currentSegment(float[] coords) {
    int type = cloudPath.pointTypes[typeIdx];
    int numCoords = curvesize[type];
    if (numCoords > 0 && affine != null) {
      affine.transform(cloudPath.pointCoords, pointIdx,
                       coords, 0,
                       numCoords / 2);
    } else {
      System.arraycopy(cloudPath.pointCoords, pointIdx, coords, 0, numCoords);
    }
    return type;
  }

  /**
   * Returns the coordinates and type of the current path segment in
   * the iteration.
   * The return value is the path segment type:
   * SEG_MOVETO, SEG_LINETO, SEG_QUADTO, SEG_CUBICTO, or SEG_CLOSE.
   * A double array of length 6 must be passed in and may be used to
   * store the coordinates of the point(s).
   * Each point is stored as a pair of double x,y coordinates.
   * SEG_MOVETO and SEG_LINETO types will return one point,
   * SEG_QUADTO will return two points,
   * SEG_CUBICTO will return 3 points
   * and SEG_CLOSE will not return any points.
   * @see PathIterator#SEG_MOVETO
   * @see PathIterator#SEG_LINETO
   * @see PathIterator#SEG_QUADTO
   * @see PathIterator#SEG_CUBICTO
   * @see PathIterator#SEG_CLOSE
   */
  public int currentSegment(double[] coords) {
    int type = cloudPath.pointTypes[typeIdx];
    int numCoords = curvesize[type];
    if (numCoords > 0 && affine != null) {
      affine.transform(cloudPath.pointCoords, pointIdx,
                       coords, 0,
                       numCoords / 2);
    } else {
      for (int i=0; i < numCoords; i++) {
        coords[i] = cloudPath.pointCoords[pointIdx + i];
      }
    }
    return type;
  }
}

