/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.util;

import java.util.ArrayList;
import java.awt.FontMetrics;

/**
 * Class used to parse strings.
 */
public class StringParser {

  /**
   * parses a string into at most maxRows strings with a maximum length
   * of maxWidth.  Used to make word wrap.
   * 
   * @param s        the string we are going to wrap
   * @param fm       the font metric used for getting the width
   * @param maxWidth the maximum allowed width
   * @param maxRows  the maximum rows we can return
   * @return an ArrayList containing the wrapped strings
   */
  public static ArrayList parseFromFront(String s, FontMetrics fm, int maxWidth, int maxRows) {

    ArrayList res = new ArrayList();
    int prevBreak, indexOfBreak;

    if (maxWidth<=0) return res; // no room for anything, return an empty ArrayList

    s = s.trim();

    while (s.length()>0 && res.size()<maxRows) {

      prevBreak = -1;
      indexOfBreak = s.indexOf(" ");

      //find the proper breaking point
      while (indexOfBreak!=-1 && fm.stringWidth(s.substring(0,indexOfBreak))<maxWidth) {
        prevBreak = indexOfBreak;
        indexOfBreak = s.indexOf(" ",indexOfBreak + 1);
      }

      if (indexOfBreak==-1 && prevBreak==-1) {
        //no more breaks in the string
        if (fm.stringWidth(s)<maxWidth) {
          res.add(new String(s));
          s = "";
        } else {
          int endIndex = s.length()-1;
          while (fm.stringWidth(s.substring(0,endIndex))>maxWidth)
            endIndex--;
          res.add(new String(s.substring(0,endIndex)));
          s = s.substring(endIndex);
        }

      } else if (indexOfBreak==-1) {
        //the last hop was to the end of the string, first check to see
        //if the whole thing will fit
        if (fm.stringWidth(s)<maxWidth) {
          res.add(new String(s));
          s = "";
        } else {
          res.add(new String(s.substring(0,prevBreak)));
          s = s.substring(prevBreak + 1);
        }
      } else if (prevBreak==-1) {
        //the first break found, at indexOfBreak, is too wide
        while (fm.stringWidth(s.substring(0,indexOfBreak))>maxWidth)
          indexOfBreak--;
        res.add(new String(s.substring(0,indexOfBreak)));
        s = s.substring(indexOfBreak);
      } else {
        //the break at prevBreak is within width
        res.add(new String(s.substring(0,prevBreak)));
        s = s.substring(prevBreak + 1);
      }

      s = s.trim();
    }

    //check to see if we have text left over, and if we do, add ... to the last line
    if (s.length()>0) {
      String lastLine = (String)res.get(res.size()-1);
      if (fm.stringWidth(lastLine + "...")<maxWidth)
        res.set(res.size()-1,lastLine + "...");
      else
        res.set(res.size()-1,lastLine.substring(0,lastLine.length()-3) + "...");
    }

    return res;
  }

}
