/**
 * Belvedere: tools for collaborative knowledge mapping
 * 
 * Copyright (C) <year> <name of author>
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere;

import java.awt.*;
import java.awt.event.*;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.ResourceBundle;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;

import belvedere.model.*;
import belvedere.util.DragDropManager;

/**
 * This class is a panel that holds a JTree used as a project/model explorer
 */
public class ProjectTreePanel extends JPanel implements ModelChangeListener,
                                                        MouseListener,
                                                        MouseMotionListener,
                                                        ProjectChangeListener {

  private JFrame parent; // kept so that we can center the dialogs
  private DragDropManager dropMan;
  private Object dragObject = null;

  private JTree tree;
  private DefaultTreeModel treeModel;
  private Project currentProj;
  private BelMutableTreeNode root;
  private String nodeString;
  private String relationString;
  private ResourceBundle props = Belvedere4.getResources();

  /**
   * Basic constructor for ProjectTreePanel
   */
  public ProjectTreePanel(JFrame parent) {
    super(new BorderLayout());
    this.parent = parent;
    this.dropMan = (DragDropManager)parent;
    this.nodeString = props.getString("NODES");
    this.relationString = props.getString("RELATIONS");
    setUpUI();
  }

  /**
   * used to set up the user interface for the ProjectTreePanel
   */
  private void setUpUI() {
    this.tree = new JTree();
    this.tree.setVisible(false);
    DefaultTreeCellRenderer renderer = new DefaultTreeCellRenderer();
    renderer.setOpenIcon(null);
    renderer.setClosedIcon(null);
    renderer.setLeafIcon(null);
    this.tree.setCellRenderer(renderer);
    JScrollPane pane = new JScrollPane(tree);
    add(pane, BorderLayout.CENTER);
    this.tree.setRootVisible(false);
    this.tree.addMouseListener(this);
    this.tree.addMouseMotionListener(this);
  }

  /**
   * called when the tree experiences a right click event, this event
   * figures out if the click occurred on a viable element and dispatches
   * the click to the proper event
   */
  private void rightClick() {
    BelMutableTreeNode mtn = (BelMutableTreeNode)tree.getLastSelectedPathComponent();
    Object selected = mtn.getUserObject();
    if (selected instanceof Model) {
      ProjectModelPropertyEditor p = new ProjectModelPropertyEditor(parent,
                                              this.props.getString("MODEL_PROPS"),
                                              (Model)selected);
      p.showEditor();
    } else if (selected instanceof Constituent) {
      mtn = (BelMutableTreeNode)tree.getSelectionPath().getPathComponent(tree.getSelectionPath().getPathCount()-3);
      Model m = (Model)mtn.getUserObject();
      ((Constituent)selected).showPropEditor(parent,m);
      // if showPropEditor caused object replace, selected will no longer be valid
    }
  }

  // current unused events of the tree
  private void singleClick() {}
  private void doubleClick() {}

  /**
   * used to set the Project for the tree
   *
   * @param prj    Project for the tree
   */
  public void setProject(Project prj) {
    this.currentProj = prj;
    this.root = new BelMutableTreeNode(currentProj);
    this.treeModel = new DefaultTreeModel(root);
    this.tree.setModel(treeModel);
  }

  /**
   * fired when a model is renamed
   *
   * @param m      model involved in rename
   */
  public void modelRenamed(Model m) {
    // renames on its own but we do have a size issue, later we may make sure
    // that after a model is renamed the tree node resizes to fit the full text
    // this can be done by removing and resinserting but with children this
    // is a problem
  }

  /**
   * this event comes from the ProjectChangeListener events, it is called when
   * the project has a model removed from it, this allows us to remove it from
   * the UI for the tree component
   *
   * @param p      Project involved in the remove
   * @param m      Model that is being removed
   * @param index  index in the array list of the model being removed
   */
  public void modelRemoved(Project p, Model m, int index) {
    BelMutableTreeNode modelNode = locateChild(root,m);
    this.treeModel.removeNodeFromParent(modelNode);
  }

  /**
   * this event comes from the ProjectChangeListener events, it is called when
   * the project has a model added to it, this allows us to add
   * the UI for the tree component for this model
   *
   * @param p      Project a model is being added to
   * @param m      Model that is being added
   */
  public void modelAdded(Project p, Model m) {
    //only one project allowed for now
    BelMutableTreeNode modelNode = new BelMutableTreeNode(m);
    this.treeModel.insertNodeInto(modelNode,root,root.getChildCount());
    // first add a node for holding vertices as children
    BelMutableTreeNode grandChild = new BelMutableTreeNode(nodeString);
    this.treeModel.insertNodeInto(grandChild,modelNode,modelNode.getChildCount());
    Constituent c;
    // add the vertices
    for (Iterator vertices = m.vertices();vertices.hasNext();) {
      c = (Constituent)vertices.next();
      this.treeModel.insertNodeInto(new BelMutableTreeNode(c),grandChild,grandChild.getChildCount());
    }
    // now a node for holding edges as children
    grandChild = new BelMutableTreeNode(relationString);
    this.treeModel.insertNodeInto(grandChild,modelNode,modelNode.getChildCount());
    // and now the edges
    for (Iterator edges = m.edges();edges.hasNext();) {
      c = (Constituent)edges.next();
      this.treeModel.insertNodeInto(new BelMutableTreeNode(c),grandChild,grandChild.getChildCount());
    }
    m.addModelChangeListener(this);
    if (this.currentProj.getModels().size()==1) {
      this.treeModel.reload();
    }
    this.tree.setVisible(true);
  }

  /**
   * When building the tree from a stored file the children of the root will not
   * display themselves properly.  Force reload makes the children visible.
   */
  public void forceReload() {
    this.treeModel.reload();
  }

  /**
   * this method is called in response to a delete event for a Constituent, this
   * comes from the ModelChangeListener
   *
   * @param m      the Model containing the Constituent that has been deleted
   * @param c      the Constituent that was deleted
   */
  public void constituentDelete(Model m, Constituent c) {
    BelMutableTreeNode parentNode;
    BelMutableTreeNode modelNode = locateChild(root,m);
    if (c instanceof Node)
      parentNode = locateChild(modelNode,nodeString);
    else
      parentNode = locateChild(modelNode,relationString);
    BelMutableTreeNode constNode = locateChild(parentNode,c);
    this.treeModel.removeNodeFromParent(constNode);
  }

  /**
   * this method is called in response to an insert event for a Constituent,
   * this comes from the ModelChangeListener
   *
   * @param m      the Model containing the Constituent that has been inserted
   * @param c      the Constituent that was inserted
   */
  public void constituentInsert(Model m, Constituent c) {
    BelMutableTreeNode modelNode = locateChild(root,m);
    BelMutableTreeNode parentNode;
    if (c instanceof Node)
      parentNode = locateChild(modelNode,nodeString);
    else
      parentNode = locateChild(modelNode,relationString);
    this.treeModel.insertNodeInto(new BelMutableTreeNode(c),parentNode,parentNode.getChildCount());
  }

  /**
   * this method is called in response to a text change event for a Constituent,
   * this comes from the ModelChangeListener
   *
   * @param m      the Model containing the Constituent whose text has changed
   * @param c      the Constituent whose text has changed
   */
  public void constituentSetText(Model m, Constituent c) {
    BelMutableTreeNode parentNode;
    BelMutableTreeNode modelNode = locateChild(root,m);
    if (c instanceof Node)
      parentNode = locateChild(modelNode,nodeString);
    else
      parentNode = locateChild(modelNode,relationString);
    BelMutableTreeNode constNode = locateChild(parentNode,c);
    //in order to get the label on node to resize had to resort to remove/add
    int position = parentNode.getIndex(constNode);
    this.treeModel.removeNodeFromParent(constNode);
    this.treeModel.insertNodeInto(new BelMutableTreeNode(c),parentNode,position);
  }

  // comes from ModelChangeListener, not used for now
  public void constituentObjectReplace(Model m, Constituent oldCons, Constituent newCons) {
    BelMutableTreeNode parentNode;
    BelMutableTreeNode modelNode = locateChild(root,m);
    if (oldCons instanceof Node)
      parentNode = locateChild(modelNode,nodeString);
    else
      parentNode = locateChild(modelNode,relationString);
    BelMutableTreeNode constNode = locateChild(parentNode,oldCons);
    //in order to get the label on node to resize had to resort to remove/add
    int position = parentNode.getIndex(constNode);
    this.treeModel.removeNodeFromParent(constNode);
    this.treeModel.insertNodeInto(new BelMutableTreeNode(newCons),parentNode,position);
  }

  // comes from ModelChangeListener, not used for now
  public void constituentSelect(Model m, Constituent c) {
    // not synchronizing for now, could use to synchronize selection
    // between other views
  }

  /**
   * the method locates the child node of parent that contains Object obj
   *
   * @param parent the node to search the children of
   * @param obj    the Object we are trying to match
   * @return returns the node with the matching obj, or null if it is not found
   */
  private BelMutableTreeNode locateChild(BelMutableTreeNode parent, Object obj) {
    Enumeration e = parent.children();
    BelMutableTreeNode child;
    while (e.hasMoreElements()) {
      child = (BelMutableTreeNode)e.nextElement();
      if (child.getUserObject()==obj) return child;
    }
    return null; // not found
  }

  /**
   * called when the mouse is pressed within the component, we use this
   * to determine dispatch to our right, single, and double click events
   *
   * @param e      associated MouseEvent
   */
  public void mousePressed(MouseEvent e) {
    int selRow = tree.getRowForLocation(e.getX(), e.getY());
    //TreePath selPath = tree.getPathForLocation(e.getX(), e.getY());
    if (selRow != -1) {
      tree.setSelectionRow(selRow);
      if (e.getModifiers()==MouseEvent.BUTTON3_MASK) {
        rightClick();
      } else if (e.getClickCount() == 1) {
        singleClick();
      } else if (e.getClickCount() == 2) {
        doubleClick();
      }
    }
  }

  /**
   * called when the mouse is released from an initial drag on this component,
   * we use this to forward the drop
   *
   * @param e      associated MouseEvent
   */
  public void mouseReleased(MouseEvent e) {
    Point p = this.tree.getLocationOnScreen();
    if (this.dragObject!=null) this.dropMan.forwardDrop((int)p.getX()+e.getX(),
                                                        (int)p.getY()+e.getY(),dragObject);
    this.dragObject = null;
  }

  /**
   * called when the mouse is dragged from an initial drag on this component, we use
   * it to initiate a drag event
   *
   * @param e      associated MouseEvent for mouseDragged
   */
  public void mouseDragged(MouseEvent e) {
    if (dragObject==null) {
      int selRow = this.tree.getRowForLocation(e.getX(), e.getY());
      //TreePath selPath = tree.getPathForLocation(e.getX(), e.getY());
      if (selRow != -1) {
        this.tree.setSelectionRow(selRow);
        BelMutableTreeNode mt = (BelMutableTreeNode)this.tree.getLastSelectedPathComponent();
        Object selected = mt.getUserObject();
        if (selected instanceof Model) {
          this.dragObject = selected;
          this.dropMan.startDrag(dragObject,e);
        } else if (selected instanceof Constituent) {
          // dragging of relations not implemented at this time
          if (selected instanceof Node) {
            this.dragObject = ((Node)selected).clone();
            this.dropMan.startDrag(dragObject,e);
          }
        }
      }
    } else {
      this.parent.getGlassPane().dispatchEvent(SwingUtilities.convertMouseEvent(this,e,this.parent.getGlassPane()));
    }
  }

  // unused mouse events
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mouseMoved(MouseEvent e) {}
  public void mouseClicked(MouseEvent e) {}

}
