/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.graphview;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

import java.util.ArrayList;
import java.util.Iterator;

import belvedere.graphview.VisualNode;
import belvedere.graphview.VisualRelation;
import belvedere.graphview.VisualConstituent;

import belvedere.model.Node;
import belvedere.model.Relation;
import belvedere.model.Constituent;
import belvedere.model.Model;
import belvedere.model.IncompatibleModelException;
import belvedere.model.ModelChangeListener;

import belvedere.util.DropTarget;
import belvedere.util.CursorFactory;
import belvedere.util.ObjectFactory;
import belvedere.util.DragDropManager;
import belvedere.util.ReverseIterator;

/**
 * this class create a JPanel that provides the "Graph" view of a model
 */
public class GraphPanel extends JPanel implements MouseMotionListener, MouseListener, ModelChangeListener {

  private Model mdl;  //reference to the contained model

  //max dimension, hard coded for now
  private Dimension size = new Dimension(1000,1000);

  private int diffX, diffY;  //diff from clk to x,y of component during drag

  //the node being dragged, if any
  private VisualNode draggedNode = null;

  //variables for dragging lines between nodes
  private VisualRelation newRelation = null;
  private boolean addingRelation = false;  //set when a relation button on ModelPanel is selected
  private String relationType; //set when the button on ModelPanel is selected
  private int newRelationDestX, newRelationDestY;

  //the node that has the hover, the reference is kept so that we can call
  //deselect() to remove the hover
  private Constituent hoverConstituent = null;

  private Cursor cursorRelation = null; //the cursor displayed in creating relation mode

  //used throughout drawing methods, faster than allocating instance var each time?
  private Iterator iter;

  protected JFrame parent;
  protected DragDropManager dropMan;
  //not to be confused with the object being dragged around within the panel
  //this is the object that is involved in a drag drop from panel to panel
  private Object dragObject = null;

  /**
   * public constructor for the GraphPanel object
   * 
   * @param mdl    the passed in model that this panel will provide the Graph
   * view of
   */
  public GraphPanel(Model mdl, JFrame parent) {
    this.mdl = mdl;
    // used to center dialogs
    this.parent = parent;
    this.dropMan = (DragDropManager)parent;
    addMouseMotionListener(this);
    addMouseListener(this);
    this.cursorRelation = CursorFactory.getCursor("../images/relcursor.gif","relation cursor",new Point(0,0),this);
  }

  /**
   * Used to set the hover to a Constituent making sure to dehover any possibly
   * hovered Constituent
   * 
   * @param c      Constituent to set the hover to
   */
  private void setHover(Constituent c) {
    if (this.hoverConstituent!=null && this.hoverConstituent!=c) 
      this.hoverConstituent.deselect();
    this.hoverConstituent = c;
    if (this.hoverConstituent!=null) this.hoverConstituent.hover();
  }

  /**
   * Invoked when the mouse has been clicked on a component, here we change
   * the selection if necessary
   * 
   * @param e      the passed in MouseEvent
   */
  public void mouseClicked(MouseEvent e) {
    int currX = e.getX();
    int currY = e.getY();
    VisualConstituent currConstituent = null;
    boolean clickedOnConstituent = false;
    //first we see if the click was on a Node
    //we do reverse so we get the one painted last...the one on top
    for (iter = this.mdl.verticesReverse();iter.hasNext();) {
      currConstituent = (Constituent)iter.next();
      if (currConstituent.withinBounds(currX,currY)) {
        clickedOnConstituent = true;
        break;
      }
    }
    //if they didn't click on a node, check to see if they click on an edge
    //once again do reverse so we get the one painted last...the one on top
    if (!clickedOnConstituent) {
      for (iter = this.mdl.edgesReverse();iter.hasNext();) {
        currConstituent = (VisualConstituent)iter.next();
        if (currConstituent.withinBounds(currX,currY)) {
          clickedOnConstituent = true;
          break;
        }
      }
    }
    if (clickedOnConstituent) {
      //whenever you click the item it goes from hover to select
      setHover(null);
      this.mdl.setSelection((Constituent)currConstituent);
      if (e.getModifiers()==MouseEvent.BUTTON3_MASK) {
        currConstituent.showPropEditor(parent, mdl);
        //if showPropEditor caused object replace, currConstituent will no longer be valid
      }
    } else { //didn't click on an element, clear the selections
      this.mdl.setSelection(null);
    }
  
  }

  /**
   * Invoked when a mouse button has been released on a component, here we
   * determine if they have released over a node to draw a new relation
   * 
   * @param e      the passed in MouseEvent
   */
  public void mouseReleased(MouseEvent e) {
    int currX = e.getX();
    int currY = e.getY();
    VisualConstituent vc = null;
    if (dragObject!=null) { //then they are doing drag/drop between frames
      Point p = getLocationOnScreen();
      this.dropMan.forwardDrop((int)p.getX()+currX,(int)p.getY()+currY,dragObject);
      this.dragObject = null;
    } else if (addingRelation && newRelation!=null) {
      //we see if the mouse was released over a Constituent
      //we do reverse so we get the one painted last...the one on top
      //first checking Nodes
      boolean releasedOnConstituent = false;
      for (iter = this.mdl.verticesReverse();iter.hasNext();) {
        vc = (VisualConstituent)iter.next();
        if (vc.withinBounds(currX,currY) && ((Relation)newRelation).getOrigin()!=vc) {
          releasedOnConstituent = true;
          break;
        }
      }
      //and now checking Edges
      if (!releasedOnConstituent) {
        for (iter = this.mdl.edgesReverse();iter.hasNext();) {
          vc = (VisualConstituent)iter.next();
          if (vc.withinBounds(currX,currY) && ((Relation)newRelation).getOrigin()!=vc) {
            releasedOnConstituent = true;
            break;
          }
        }
      }
      //and now handle the release if it indicates drawing a Relation
      if (releasedOnConstituent) {
        ((Relation)newRelation).setDestination((Constituent)vc);
        this.mdl.insertEdge((Relation)newRelation);
        //we change the selection to the destination end of the new relation
        //this takes care of a hover problem between the sister panels
        setHover(null);
        this.mdl.setSelection((Constituent)vc);
      }
      newRelation = null; //now set up to start making another
      repaint();
    }
    // this is done to make sure listeners will repaint given drag events
    this.mdl.setSelection(mdl.getSelection());
  }

  /**
   * Invoked when a mouse button is pressed on a component and then
   * dragged.  Mouse drag events will continue to be delivered to
   * the component where the first originated until the mouse button is
   * released (regardless of whether the mouse location is within the
   * bounds of the component).  Here we handle either the creation of a new
   * relation or the dragging of Nodes
   * 
   * @param e      the passed in MouseEvent
   */
  public void mouseDragged(MouseEvent e) {
    int currX = e.getX();
    int currY = e.getY();
    VisualConstituent vc = null;
    if (addingRelation) {
      if (newRelation==null) {  //determine if we have start of making relation
        boolean startingNewRelation = false;
        //start with the Nodes
        //we do reverse so we get the one painted last...the one on top
        for (iter = this.mdl.verticesReverse();iter.hasNext();) {
          vc = (VisualConstituent)iter.next();
          if (vc.withinBounds(currX,currY)) {
            startingNewRelation = true;
            break;
          }
        }
        //if not dragging from a Node, now check the Edges
        if (!startingNewRelation) {
          //we do reverse so we get the one painted last...the one on top
          for (iter = this.mdl.edgesReverse();iter.hasNext();) {
            vc = (VisualConstituent)iter.next();
            if (vc.withinBounds(currX,currY)) {
              startingNewRelation = true;
              break;
            }
          }
        }
        //and now handle it if the start of drag indicates making new Relation
        if (startingNewRelation) {
          newRelation = (VisualRelation)ObjectFactory.createObject(relationType);
          ((Relation)newRelation).setOrigin((Constituent)vc);
          newRelationDestX = currX;
          newRelationDestY = currY;
          setHover(null);
        }
        if (newRelation==null) {
          // in relation adding mode but dragging on empty canvas
          setHover(null);
          this.mdl.setSelection(null);
        }
      } else { //newRelation is not null, so we r just drawing newRelation to the x,y
        setHover(null);
        boolean needHover = false;
        //we see if we should be hovering a Consituent, first check Nodes
        //we do reverse so we get the one painted last...the one on top
        for (iter = this.mdl.verticesReverse();iter.hasNext();) {
          vc = (VisualConstituent)iter.next();
          if (vc.withinBounds(currX,currY)) {
            needHover = true;
            break;
          }
        }
        //and now check the Edges if a Vertex shouldn't have the hover
        if (!needHover) {
          for (iter = this.mdl.edgesReverse();iter.hasNext();) {
            vc = (VisualConstituent)iter.next();
            if (vc.withinBounds(currX,currY)) {
              needHover = true;
              break;
            }
          }
        }
        //and now handle the hover if needed
        if (needHover)
          if (vc!=this.mdl.getSelection()) setHover((Constituent)vc);
        this.newRelationDestX = currX;
        this.newRelationDestY = currY;
      }
    } else {  //drag not in adding a relation mode
      if (dragObject!=null) { //they are in drag drop state
        this.parent.getGlassPane().dispatchEvent(SwingUtilities.convertMouseEvent(this,e,this.parent.getGlassPane()));
        // don't let it get down to that setSelection that forces a repaint
        // only the glassPane needs to repaint so we return so that these
        // events are fired faster
        return;
      } else { //they are dragging a Node around in drag drop state
        //don't let them drag off left or top
        if (currX<0) currX=0;
        if (currY<0) currY=0;
        if (currX>size.getWidth()) currX = (int)size.getWidth();
        if (currY>size.getHeight()) currY = (int)size.getHeight();
        if (draggedNode!=null) {
          this.draggedNode.setX(currX - diffX);
          this.draggedNode.setY(currY - diffY);
        } else {  //check to see if this is the beginning of a drag/drop
          boolean draggingNode = false;
          //we do reverse so we get the one painted last...the one on top
          for (iter = this.mdl.verticesReverse();iter.hasNext();) {
            vc = (VisualConstituent)iter.next();
            if (vc.withinBounds(currX,currY)) {
              if (e.isShiftDown()) { //start of drag/drop
                this.dragObject = ((Node)vc).clone();
                this.dropMan.startDrag(dragObject,e);
              } else { //start of regular node dragging...not drag drop frame to frame
                draggingNode = true;
                diffX = currX - ((Node)vc).getX();
                diffY = currY - ((Node)vc).getY();
                this.draggedNode = (Node)vc;
                //don't let it have hover and selection at the same time
                //otherwise losing the hover will deselect
                setHover(null);
                this.mdl.setSelection((Constituent)vc);
              }
              break;
            }
          }
          if (!draggingNode) {  //dragging on empty canvas
            setHover(null);
            this.mdl.setSelection(null);
            //repaint(); not necessary?  forced repaint below
          }
        }
      }
    }
    // force a repaint in all listeners
    this.mdl.setSelection(mdl.getSelection());
  }

  /**
   * Invoked when the mouse button has been moved on a component
   * (with no buttons down), here we handle hovering.
   * 
   * @param e      the passed in MouseEvent
   */
  public void mouseMoved(MouseEvent e) {
    draggedNode = null; //not dragging a node around
    int currX = e.getX();
    int currY = e.getY();
    VisualConstituent vc = null;
    //this method only sets the hover, exits when found and set
    //we do reverse so we get the one painted last...the one on top
    for (iter = this.mdl.verticesReverse();iter.hasNext();) {
      vc = (VisualConstituent)iter.next();
      if (vc.withinBounds(currX,currY)) {
        // don't hover the selection
        if (this.hoverConstituent!=vc && vc!=this.mdl.getSelection())
          setHover((Constituent)vc);
        repaint();
        return;
      }
    }
    //didn't hover a Vertex? then handle the Edges
    //we do reverse so we get the one painted last...the one on top
    for (iter = this.mdl.edgesReverse();iter.hasNext();) {
      vc = (VisualConstituent)iter.next();
      if (vc.withinBounds(e.getX(),e.getY())) {
        // don't hover the selection
        if (this.hoverConstituent!=vc && vc!=this.mdl.getSelection())
          setHover((Constituent)vc);
        repaint();
        return;
      }
    }
    //if it got here it wasn't over a constituent, so if one was hovered,
    //deselect it
    if (this.hoverConstituent!=null) {
      setHover(null);
      repaint();
    }
  }

  //here because we are implementing MouseMotionListener, not used for now
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void mousePressed(MouseEvent e) {}
  
  /**
   * paints the surface of the GraphPanel
   * 
   * @param g      the Graphics to paint upon
   */
  public void paint(Graphics g) {
    VisualConstituent vc;
    //draw the background white
    g.setColor(Color.white);
    g.fillRect(0,0,getWidth(),getHeight());
    g.setColor(Color.black);
    //now iterate through and draw the edges
    //drawing reverse so that edges to edges don't overlap previously drawn edge
    for (iter = this.mdl.edgesReverse();iter.hasNext();) {
      vc = (VisualConstituent)iter.next();
      vc.paint(g);
    }
    //iterate through and draw the vertices
    for (iter = this.mdl.vertices();iter.hasNext();) {
      vc = (VisualConstituent)iter.next();
      vc.paint(g);
    }
    //draw the selected Constituent, if any, on top
    VisualConstituent selected = this.mdl.getSelection();
    if (selected!=null && selected instanceof VisualNode) selected.paint(g);
    if (this.newRelation!=null) this.newRelation.paint(g,this.newRelationDestX,this.newRelationDestY);
  }

  /**
   * method used to add a new Node to the model and thus the GraphPanel
   * 
   * @param nodeType the type of node to add
   * @param text     the text for the new Node object
   */
  public void addNode(String nodeType, String text) {
    this.mdl.insertNode(nodeType, text);
  }

  /**
   * sets the GraphPanel to be in or out of the adding relation mode
   * 
   * @param relationType
   *               the type of relation to add when adding a relation
   * @param b      true to put the panel in adding relation mode, false to
   *               quit adding relation mode
   */
  public void addRelationMode(String relationType, boolean b) {
    this.addingRelation = b;
    this.relationType = relationType;
    if (addingRelation)
      setCursor(this.cursorRelation);
    else
      setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
  }

  //these are used to keep the size at a preset so scrolling can be used
  public Dimension getMinimumSize() { return size;}
  public Dimension getMaximumSize() { return size;}
  public Dimension getPreferredSize() { return size;}

  /**
   * this method is called by the application DragDropManager when
   * a drop has occurred on this component
   * 
   * @param x      x coordinate of drop
   * @param y      y coordinate of drop
   * @param o      Object associated with drop
   */
  public void drop(int x, int y, Object o) {
    if (o instanceof Model) {
      if (o==this.mdl) return;
      try {
        this.mdl.addModel((Model)o);
      } catch (IncompatibleModelException e) {

      }
    } else if (o instanceof Node) {
      Point p = getLocationOnScreen();
      int dropX = x - (int)p.getX();
      int dropY = y - (int)p.getY();
      Node newNode = (Node)((Node)o).clone();
      this.mdl.insertVertex(newNode);
      newNode.setX(dropX);
      newNode.setY(dropY);
    } else if (o instanceof Relation) {
      // code here for Relation drop...not implemented
    }
  }

  /**
   * fired when the model associated with this GraphPanel is renamed
   * 
   * @param m      model involved in rename, always the associated model
   */
  public void modelRenamed(Model m) {}

  /**
   * fired when a Constituent is deleted
   * 
   * @param m      model involved in the delete
   * @param p      Constituent deleted
   */
  public void constituentDelete(Model m, Constituent c) {
    repaint();
  }

  /**
   * fired when a Constituent is inserted
   * 
   * @param m      model involved in the insert
   * @param c      Constituent inserted
   */
  public void constituentInsert(Model m, Constituent c) {
    repaint();
  }

  /**
   * fired when a Constituent's text is changed
   * 
   * @param m      model involved in the text change
   * @param c      Constituent involved in text change
   */
  public void constituentSetText(Model m, Constituent c) {
    repaint();
  }


  /**
   * fired when a Constituent is switched to a different object type
   * as in the Evidence Model
   * 
   * @param m      model involved in the replace
   * @param c      Constituent involved in the object swap
   */
  public void constituentObjectReplace(Model m, Constituent oldCons, Constituent newCons) {
    repaint();
  }

  /**
   * fired when the selection in the model changes
   * 
   * @param m      the model that has had a selection change
   * @param c      the newly selected Constituent, possibly null
   */
  public void constituentSelect(Model m, Constituent c) {
    repaint();
  }

}
