/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import java.lang.reflect.*;

import belvedere.util.ButtonDescriptor;
import belvedere.util.ObjectFactory;
import belvedere.util.ReverseIterator;

import belvedere.model.concept.*;

/**
 * A model represents a collection of nodes and relations
 * among these nodes.
 */
public class Model extends BelvedereObject {

  protected ArrayList vertices = new ArrayList();
  protected ArrayList edges = new ArrayList();

  private ArrayList modelChangeListeners = new ArrayList();

  // set equal to "" to avoid null situations later
  protected String name = "";
  protected String creator = "";
  protected String creationDate = "";
  protected String description = "";

  protected boolean dirty = false;  //dirty if changes made since last save

  protected Constituent selConstituent = null;

  // sub classes complain if this is not here?
  public Model() {
    this.name = "Simple Model";
  }

  /**
   * Constructor to create a model with a specific name.
   *
   * @param name  Name for Model being created.
   */
  public Model(String name) {
    this.name = name;
  }

  /**
   * adds an object implementing the ModelChangeListener interface to the
   * list of listeners notified of model change events
   * 
   * @param mc     the ModelChangeListener to add to the list
   */
  public void addModelChangeListener(ModelChangeListener mc) {
    modelChangeListeners.add(mc);
  }

  /**
   * Sets the name for this model.
   *
   * @param name  name this model is to be set to.
   */
  public void setName(String name) {
    this.name = name;
    fireModelRenamed();
  }

  /**
   * Returns the name of this model.
   *
   * @return name  String name of this model.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Returns the Creator of this project.
   *
   * @return Creator  Creator of this project
   */
  public String getCreator() {
    return this.creator;
  }

  /**
   * Sets the creator of this project.
   *
   * @param creator   Text String naming this project
   */
  public void setCreator(String creator) {
    this.creator = creator;
  }

  /**
   * Returns the creationDate of this project.
   *
   * @return creationDate  creationDate of this project
   */
  public String getCreationDate() {
    return this.creationDate;
  }

  /**
   * Sets the creationDate of this project.
   *
   * @param creationDate   Text String for creationDate of project
   */
  public void setCreationDate(String creationDate) {
    this.creationDate = creationDate;
  }

  /**
   * Returns the description of this project.
   *
   * @return description  description of this project
   */
  public String getDescription() {
    return this.description;
  }

  /**
   * Sets the description of this project.
   *
   * @param description   Text String naming this project
   */
  public void setDescription(String description) {
    this.description = description;
  }

  /**
   * sets the dirty flag for this model
   * 
   * @param d      sets the dirty flag to d = true or d = false
   */
  public void setDirty(boolean d) {
    this.dirty = d;
  }

  /**
   * used to determine if the model is dirty
   * 
   * @return true if dirty, false otherwise
   */
  public boolean isDirty() {
    return this.dirty;
  }

  /**
   * used to set the text of a constituent in the model.  We use this method instead
   * of doing it directly so the appropriate ModelChangeListener event can be fired.
   * 
   * @param c      Constituent to set the text of
   * @param s      new text for Constituent
   */
  public void setConstituentText(Constituent c, String s) {
    c.setText(s);
    fireConstituentSetText(c);
  }

  public Constituent getSelection() {
    return this.selConstituent;
  }

  public void setSelection(Constituent c) {
    if (this.selConstituent!=null && this.selConstituent!=c)
      this.selConstituent.deselect();
    this.selConstituent = c;
    if (this.selConstituent!=null) this.selConstituent.select();
    fireSelectionEvent();
  }

  /**
   * used to insert a new relation into the model
   * 
   * @param vp        one endpoint of the new relation
   * @param wp        other endpoint of the new relation
   * @param className type of relation to create
   * @return newly inserted Constituent
   * @exception InvalidPositionException
   */
  public Constituent insertRelation(Constituent oc,Constituent dc, String className) /* e */ {
    //don't allow multiple relations between nodes in this model
    if (areAdjacent(oc,dc)) return null;
    Class relationDefinition;
    Class stringArg[] = new Class[] {String.class, String.class, String.class,
      String.class, String.class, int.class};
    Object args[] = new Object[] {"","","","","",new Integer(Constituent.NEUTRALLY_BELIEVE)};
    Constructor relationConstructor;
    Relation relation;

    try {
      relationDefinition = Class.forName(className);
      relationConstructor = 
      relationDefinition.getConstructor(stringArg);
      relation = (Relation)ObjectFactory.createObjectWithArgs(relationConstructor, args);
    } catch (ClassNotFoundException e) {
      relation = null;
    } catch (NoSuchMethodException e) {
      relation = null;
    }


    if (relation!=null) {
      this.dirty = true;
      this.edges.add(relation);
      fireConstituentInsert(relation);
      return relation;
    }

    return null;

  }

  /**
   * used to insert a new node into a model via the class name
   * 
   * @param className class of node to create
   * @param text      text for the new node
   * @return newly inserted Constituent object
   */
  public Constituent insertNode(String className, String text) {

    Class nodeDefinition;
    Class stringArg[] = new Class[] {String.class, String.class, String.class,
      String.class, String.class, int.class,
      int.class, int.class, int.class};
    Object args[] = new Object[] {text,"","","","",new Integer(Constituent.NEUTRALLY_BELIEVE),
      new Integer(0),new Integer(0),
      new Integer(200)};
    Constructor nodeConstructor;
    Node node;

    try {
      nodeDefinition = Class.forName(className);
      nodeConstructor = 
      nodeDefinition.getConstructor(stringArg);
      node = (Node)ObjectFactory.createObjectWithArgs(nodeConstructor, args);
    } catch (ClassNotFoundException e) {
      node = null;
    } catch (NoSuchMethodException e) {
      node = null;
    }

    if (node!=null) {
      this.dirty = true;
      this.vertices.add(node);
      fireConstituentInsert(node);
      return node;
    }

    return null;

  }

  /**
   * used to insert a new node into the model
   * 
   * @param n      the node to insert
   */
  public void insertVertex(Node n) {
    this.vertices.add(n);
    fireConstituentInsert(n);
  }

  /**
   * used to insert a new relation into the model
   * 
   * @param r      the relation to insert
   */
  public void insertEdge(Relation r) {
    //basic models reject multiple edges, override for multiple edges
    if (this.areAdjacent(r.getOrigin(),r.getDestination())) return;
    this.edges.add(r);
    fireConstituentInsert(r);
  }

  /**
   * used to delete the selected constituent from the model
   */
  public void deleteSelected() {
    if (this.selConstituent!=null) {
      //this could be consolidated into one method, removeConstituent, however
      //some model types must do some extra bookeeping when removing Edges
      if (this.selConstituent instanceof Node) {
        removeVertex((Node)this.selConstituent);
      } else {  //VisualRelation
        removeEdge((Relation)this.selConstituent);
      }
    }
  }

  /**
   * removes a vertex and thus a node from the model
   * 
   */
  public void removeVertex(Node n) {
    this.dirty = true;
    if (n==selConstituent) {
      this.selConstituent = null;
      fireSelectionEvent();
    }
    //remove all edges connected to this node
    for (int i = edges.size() - 1; i > -1;i--) {
      Relation currRelation = (Relation)edges.get(i);
      if (currRelation.getOrigin()==n || currRelation.getDestination()==n)
        removeEdge(currRelation);
    }
    for (int i = vertices.size() - 1;i > -1;i--) {
      Node currNode = (Node)this.vertices.get(i);
      if (currNode==n) {
        this.vertices.remove(i);
        break;
      }
    }
    fireConstituentDelete(n);
  }

  /**
   * removes an edge and thus a relation from the model
   * 
   * @param ep     Position holding the relation to remove
   * @exception InvalidPositionException
   */
  public void removeEdge(Relation r) {
    this.dirty = true;
    if (r==this.selConstituent) {
      this.selConstituent = null;
      fireSelectionEvent();    
    }
    //remove all edges connected to this edge, note that only edges added after
    //an edge can be connected, so no danger of deleting lower array positions
    //out from under this decrementing loop
    for (int i = edges.size() - 1; i > -1;i--) {
      Relation currRelation = (Relation)edges.get(i);
      if (currRelation.getOrigin()==r || currRelation.getDestination()==r)
        this.removeEdge(currRelation);
    }
    //remove the edge
    for (int i = edges.size() - 1; i > -1;i--) {
      Relation currRelation = (Relation)edges.get(i);
      if (currRelation==r) {
        edges.remove(i);
        break;
      }
    }
    fireConstituentDelete(r);
  }

  public Iterator vertices() {
    return this.vertices.iterator();
  }

  public Iterator edges() {
    return this.edges.iterator();
  }

  public boolean areAdjacent(Constituent a, Constituent b) /* e */ {
    for (Iterator i = edges.iterator();i.hasNext();) {
      Relation r = (Relation)i.next();
      if ((r.getOrigin()==a && r.getDestination()==b) ||
          (r.getOrigin()==b && r.getDestination()==a)) return true;
    }
    return false;
  }

  public boolean areDirectedAdjacent(Constituent oc, Constituent dc) {
    for (Iterator i = edges.iterator();i.hasNext();) {
      Relation r = (Relation)i.next();
      if (r.getOrigin()==oc && r.getDestination()==dc)return true;
    }
    return false;
  }

  public Relation getDirectedAdjoiningEdge(Constituent oc, Constituent dc) {
    for (Iterator i = edges.iterator();i.hasNext();) {
      Relation r = (Relation)i.next();
      if (r.getOrigin()==oc && r.getDestination()==dc)return r;
    }
    return null;
  }

  public Constituent[] endVertices(Relation r) /* e */ {
    Constituent cons[] = new Constituent[2];
    cons[0] = r.getOrigin();
    cons[1] = r.getDestination();
    return cons;
  }

  public Iterator verticesReverse() {
    //don't know how to do yet!
    //return vertices.iterator();
    return new ReverseIterator(vertices);
  }

  public Iterator edgesReverse() {
    //don't know how to do yet!
    //return edges.iterator();
    return new ReverseIterator(edges);
  }

  /**
   * used to notify the ModelChangeListeners of a model renamed event
   * 
   */
  protected void fireModelRenamed() {
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).modelRenamed(this);
  }

  /**
   * used to notify the ModelChangeListeners of a constituent insert event
   * 
   * @param c      contituent inserted
   */
  protected void fireConstituentInsert(Constituent c) {
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).constituentInsert(this, c);
  }

  /**
   * used to notify the ModelChangeListeners of a constituent delete event
   * 
   * @param c      contituent deleted
   */
  protected void fireConstituentDelete(Constituent c) {
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).constituentDelete(this, c);
  }

  /**
   * used to notify the ModelChangeListeners of a constituent set text event
   * 
   * @param c      contituent inserted
   */
  protected void fireConstituentSetText(Constituent c) {
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).constituentSetText(this, c);
  }

  /**
   * used to notify the ModelChangeListeners of a constituent replace
   * 
   * @param oldCons constituent being replaced
   * @param newCons constituent replacing the oldCons
   */
  protected void fireConstituentObjectReplace(Constituent oldCons, Constituent newCons) {
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).constituentObjectReplace(this,oldCons,newCons);
  }

  /**
   * used to notify the ModelChangeListeners of a constituent selection event
   */
  private void fireSelectionEvent() {
    //would have liked to put this is mouseReleased, but it was out of sync
    //with the selection in Click and Drag
    for (int i=0;i<modelChangeListeners.size();i++)
      ((ModelChangeListener)modelChangeListeners.get(i)).constituentSelect(this,selConstituent);
  }

  /**
   * this method provides an array list of ButtonDescriptors describing the
   * buttons that should be created to work with the type of Nodes that this
   * Model uses
   * 
   * @return an ArrayList of ButtonDescriptor objects
   */
  public ArrayList getNodeButtonDescriptors() {
    ArrayList al = new ArrayList();
    ButtonDescriptor bd = new ButtonDescriptor("belvedere.model.Node","../images/addnode.gif","Add Node");
    al.add(bd);
    return al;
  }

  /**
   * this method provides an array list of ButtonDescriptors describing the
   * buttons that should be created to work with the type of Relations that this
   * Model uses
   * 
   * @return an ArrayList of ButtonDescriptor objects
   */
  public ArrayList getRelationButtonDescriptors() {
    ArrayList al = new ArrayList();
    ButtonDescriptor bd = new ButtonDescriptor("belvedere.model.Relation","../images/addrelation.gif","Add Relation");      
    al.add(bd);
    return al;
  }

  /**
   * overriding toString() to return name field
   * 
   * @return name field of Model
   */
  public String toString() {
    return this.name;
  }

  /**
   * used to replace a vertex in the model with another
   * 
   * @param oldNode vertex being replaced
   * @param newNode replacement vertex
   */
  public void replaceVertex(Node oldNode, Node newNode) /* e */ {
    //you must switch the edges to the new node BEFORE deleting old
    for (Iterator iter = edges.iterator();iter.hasNext();) {
      Relation r = (Relation)iter.next();
      if (r.getOrigin()==oldNode) r.setOrigin(newNode);
      if (r.getDestination()==oldNode) r.setDestination(newNode);
    }
    //notice that we aren't using removeVertex, we replace manually so that oldNode
    //will still be present for fireConstituentObjectReplace and also we put it
    //back in the same array location so the layering is the same
    int i;
    for (i = 0;vertices.get(i)!=oldNode;i++) ;
    this.vertices.remove(i);
    this.vertices.add(i,newNode);
    fireConstituentObjectReplace(oldNode,newNode);
  }

  /**
   * used to replace a edge in the model with another
   * 
   * @param oldRelation edge being replaced
   * @param newRelation replacement edge
   */  
  public void replaceEdge(Relation oldRelation, Relation newRelation) /* e */ {
    //you must switch the edges to the new edge BEFORE deleting old
    for (Iterator iter = edges.iterator();iter.hasNext();) {
      Relation r = (Relation)iter.next();
      if (r.getOrigin()==oldRelation) r.setOrigin(newRelation);
      if (r.getDestination()==oldRelation) r.setDestination(newRelation);
    }
    //notice that we aren't using removeEdge, we replace manually so that oldRelation
    //will still be present for fireConstituentObjectReplace and also we put it
    //back in the same array location so the layering is the same, and when saving
    //any edge to edge ordering that needs to be present in the XML is maintained
    int i;
    for (i = 0;edges.get(i)!=oldRelation;i++) ;
    this.edges.remove(i);
    this.edges.add(i,newRelation);
    fireConstituentObjectReplace(oldRelation,newRelation);
  }

  /**
   * Adds all components of <code>Model</code> parameter to this <code>Model</code>.
   *
   * @param model  <code>Model</code> to be combined with this <code>Model</code>
   */
  public void addModel(Model mergingModel) throws IncompatibleModelException {
    Iterator vertices, edges;
    HashMap newVertices = new HashMap();
    Constituent oldCons, sourceCons, destCons;
    Relation nextEdge;
    Node nextNode;
    Relation nextRelation;
    vertices = mergingModel.vertices();
    edges = mergingModel.edges();
    try {
      while (vertices.hasNext()) {
        oldCons = (Constituent)vertices.next();
        nextNode = (Node)((Node)oldCons).clone();
        insertVertex(nextNode);
        newVertices.put(oldCons, nextNode);
      }
      while (edges.hasNext()) {
        nextEdge = (Relation)edges.next();
        nextRelation = (Relation)nextEdge.clone();
        sourceCons = (Constituent)newVertices.get(nextEdge.getOrigin());
        destCons = (Constituent)newVertices.get(nextEdge.getDestination());
        nextRelation.setOrigin(sourceCons);
        nextRelation.setDestination(destCons);
        insertEdge(nextRelation);
      }
    } 
    // this should not occur, but there is no guarantee that a model will
    // contain Nodes and Relations 
    catch (ClassCastException e) {
      throw new IncompatibleModelException("Unable to merge models.");
    }
  }

}
