/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;

import java.util.HashMap;

import belvedere.graphview.VisualRelation;

/**
 * the basic relation object from which all other relations should extend
 */
public class Relation extends Constituent implements VisualRelation {

  protected Constituent origin;
  protected Constituent destination;

  /**
   * Constructor for Relation
   */
  public Relation() {
    super("","","","","",Constituent.NEUTRALLY_BELIEVE);
    //set up current as regular
    this.borderColor = this.regBorderColor;
    this.borderStroke = this.regBorderStroke;
  }

  /**
   * Constructor for Relation
   * 
   * @param text     text for the node
   * @param URL      URL for the node
   * @param notes    notes for the node
   * @param creator  creator of the node
   * @param modifier modifier of the node
   * @param strength strength of the node, indicated in line width
   */
  public Relation(String text, String URL, String notes, 
                  String creator, String modifier, int strength) {
    super(text,URL,notes,creator,modifier,strength);
    //set up current as regular
    this.borderColor = this.regBorderColor;
    this.borderStroke = this.regBorderStroke;
  }

  /**
   * used to create a HashMap containing the
   * attributes to be saved in the XML representation
   * 
   * @return populated HashMap
   */
  public HashMap getAttributes() {
    HashMap h = super.getAttributes();
    h.put(Project.ORIGIN_GUID_KEY,this.origin.getGUID());
    h.put(Project.DEST_GUID_KEY,this.destination.getGUID());
    return h;
  }

  /**
   * Returns a new <code>Relation</code> object which is a copy of this instance.
   *
   * @return clone  a copy <code>Object</code> instance of this 
   *               <code>Relation</code>.
   */
  public Object clone() {
    return new Relation(this.text, this.URL, this.notes, this.creator, 
                        this.modifier, this.strength);
  }

  /**
   * returns the origin Constituent for a relation
   * 
   * @return the origin Constituent of the relation
   */
  public Constituent getOrigin() {
    return this.origin;
  }

  /**
   * sets the origin Constituent for a Relation
   * 
   * @param c      the new origin Constituent for the Relation
   */
  public void setOrigin(Constituent c) {
    this.origin = c;
  }

  /**
   * returns the destination Constituent for a relation
   * 
   * @return the destination Constituent of the relation
   */
  public Constituent getDestination() {
    return this.destination;
  }

  /**
   * sets the destination Constituent for a Relation
   * 
   * @param c      the new destination Constituent for the Relation
   */
  public void setDestination(Constituent c) {
    this.destination = c;
  }

  //sensitivity for withinBounds(x,y) checks using a slope comparison formula
  //the derived class may wish to change the value to accomodate thicker lines
  public static int sensitivity = 3000;

  /**
   * Used to paint this Relation
   * 
   * @param g      the graphics to paint the line to
   */
  public void paint(Graphics g) {
    int x1 = this.getOrigin().getCenterX();
    int y1 = this.getOrigin().getCenterY();
    int x2 = this.getDestination().getCenterX();
    int y2 = this.getDestination().getCenterY();
    Graphics2D g2D = (Graphics2D)g;
    g2D.setColor(borderColor);
    g2D.setStroke(borderStroke);
    g2D.drawLine(x1,y1,x2,y2);
  }

  /**
   * Used to paint a relation that is being added to a model
   * that is it has the origin end set to a constituent and
   * the destination is still being dragged out to x2, y2
   * 
   * @param g      the graphics to paint the line to
   */
  public void paint(Graphics g, int x2, int y2) {
    int x1 = this.getOrigin().getCenterX();
    int y1 = this.getOrigin().getCenterY();
    Graphics2D g2D = (Graphics2D)g;
    g2D.setColor(borderColor);
    g2D.setStroke(borderStroke);
    g2D.drawLine(x1,y1,x2,y2);
  }

  /**
   * returns the horizontal center of the constituent
   * 
   * @return horizontal center of the constituent
   */
  public int getCenterX() {
    return (this.origin.getCenterX() + this.destination.getCenterX()) / 2;
  }

  /**
   * returns the vertical center of the constituent
   * 
   * @return vertical center of the constituent
   */    
  public int getCenterY() {
    return (this.origin.getCenterY() + this.destination.getCenterY()) / 2;
  }

  /**
   * This function determines if x,y is a click on the line from x1,y2 to y1,y2.
   * 
   * @param x      x coordinate that we are testing
   * @param y      y coordinate that we are testing
   * @return true if x,y is withinBounds of the line, false otherwise
   */
  public boolean withinBounds(int x, int y) {
    //determine endpoints
    int x1 = this.origin.getCenterX();
    int y1 = this.origin.getCenterY();
    int x2 = this.destination.getCenterX();
    int y2 = this.destination.getCenterY();
    //proportional slope test to the two endpoints
    if (Math.abs((y - y1)*(x2 - x) - (x - x1)*(y2 - y)) < this.sensitivity) {
      //now make sure the click was between the endpoints
      if (x1<=x2) {
        if (y1<=y2) {
          return((x>=x1 && x<=x2) && (y>=y1 && y<=y2));
        } else { // y1>y2
          return((x>=x1 && x<=x2) && (y<=y1 && y>=y2));
        }
      } else { // x1>x2
        if (y1<=y2) {
          return((x<=x1 && x>=x2) && (y>=y1 && y<=y2));
        } else { // y1>y2
          return((x<=x1 && x>=x2) && (y<=y1 && y>=y2));
        }
      }
    }
    return false;
  }
}
