/**
 * Belvedere: tools for collaborative knowledge mapping 
 *
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model.concept;

import java.lang.reflect.*;
import java.util.ArrayList;
import java.util.Enumeration;

import belvedere.model.*;
import belvedere.util.ButtonDescriptor;
import belvedere.util.ObjectFactory;

/**
 * A model represents a collection of nodes and relations
 * among these nodes.
 */
public class ConceptualModel extends Model {

  /**
   * Basic constructor
   */
  public ConceptualModel() {
    this.name = "Conceptual Model";
  }

  /**
   * Constructor with model name
   * 
   * @param name   name of the model
   */
  public ConceptualModel(String name) {
    this.name = name;
  }

  /**
   * used to add a relation to a conceptual model, we are overriding this
   * because Conceptual Models allow two relations between nodes, also, they
   * keep track of relationCount so that only the first relation in drawing
   * order draws itself
   * 
   * @param vp
   * @param wp
   * @param className
   * @return 
   * @exception InvalidPositionException
   */
  public Constituent insertRelation(Constituent oc, Constituent dc, String className) /* e */ {

    boolean areDirectedAdjacent = this.areAdjacent(oc,dc);      

    if (areDirectedAdjacent) return null;

    Class relationDefinition;
    Class stringArg[] = new Class[] {String.class, String.class, String.class,
      String.class, String.class, int.class};
    Object args[] = new Object[] {"","","","","",new Integer(Constituent.NEUTRALLY_BELIEVE)};
    Constructor relationConstructor;
    ConceptualRelation relation;

    try {
      relationDefinition = Class.forName(className);
      relationConstructor = 
      relationDefinition.getConstructor(stringArg);
      relation = (ConceptualRelation)ObjectFactory.createObjectWithArgs(relationConstructor, args);
    } catch (ClassNotFoundException e) {
      relation = null;
    } catch (NoSuchMethodException e) {
      relation = null;
    }

    if (relation!=null) {
      dirty = true;
      if (areDirectedAdjacent(dc,oc))
        relation.setRelCount(1);
      else
        relation.setRelCount(0);
      // must be directed so we can use that information to determine if
      // there are already edges from vp -> wp and vice versa
      relation.setOrigin(oc);
      relation.setDestination(dc);
      insertEdge(relation);
      fireConstituentInsert(relation);
      return relation;
    }

    return null;

  }

  /**
   * used to add a relation to a conceptual model, we are overriding this
   * because Conceptual Models allow two relations between nodes, also, they
   * keep track of relationCount so that only the first relation in drawing
   * order draws itself.  Note that this is the method that is called when
   * restoring from XML.
   * 
   * @param vp
   * @param wp
   * @param className
   * @return 
   * @exception InvalidPositionException
   */
  public void insertEdge(Relation r) /* e */ {
    //this model allows an edge in each direction, only check directed
    if (this.areDirectedAdjacent(r.getOrigin(),r.getDestination())) return;
    //number the edge for proper painting of the triangle wedge
    if (this.areDirectedAdjacent(r.getDestination(),r.getOrigin()))
      ((ConceptualRelation)r).setRelCount(1);
    else
      ((ConceptualRelation)r).setRelCount(0);
    edges.add(r);
  }

  /**
   * used to remove an edge from a conceptual mode, we override removeEdge so we can
   * renumber the remaining edge so that it will draw its line again
   * 
   * @param ep     the edge position to remove
   * @exception InvalidPositionException
   */
  public void removeEdge(ConceptualRelation cr) /* e */ {
    //get the endpoints to determine if still connected after delete
    Constituent cons[] = this.endVertices(cr);
    super.removeEdge(cr);
    //modl allows 2 edges, at most one following will be true after above delete
    if (areDirectedAdjacent(cons[0],cons[1])) {
      ConceptualRelation opposite = (ConceptualRelation)this.getDirectedAdjoiningEdge(cons[0],cons[1]);     
      opposite.setRelCount(0);
    } else if (areDirectedAdjacent(cons[1],cons[0])) {
      ConceptualRelation opposite = (ConceptualRelation)this.getDirectedAdjoiningEdge(cons[1],cons[0]);
      opposite.setRelCount(0);
    }
  }

  /**
   * this method provides an array list of ButtonDescriptors describing the
   * buttons that should be created to work with the type of Nodes that this
   * Model uses
   * 
   * @return an ArrayList of ButtonDescriptor objects
   */
  public ArrayList getNodeButtonDescriptors() {
    ArrayList al = new ArrayList();
    ButtonDescriptor bd = new ButtonDescriptor("belvedere.model.concept.ConceptualNode","../images/addnode.gif","Add Node");
    al.add(bd);
    return al;
  }

  /**
   * this method provides an array list of ButtonDescriptors describing the
   * buttons that should be created to work with the type of Relations that this
   * Model uses
   * 
   * @return an ArrayList of ButtonDescriptor objects
   */
  public ArrayList getRelationButtonDescriptors() {
    ArrayList al = new ArrayList();
    ButtonDescriptor bd = new ButtonDescriptor("belvedere.model.concept.ConceptualRelation","../images/addconceptualrelation.gif","Add Relation");
    al.add(bd);
    // just for testing
    //bd = new ButtonDescriptor("belvedere.model.concept.ConceptualRelation","../images/addrelation.gif","Add Conceptual Relation");
    //al.add(bd);
    return al;
  }


  /**
   * Adds all components of <code>Model</code> parameter to this 
   * <code>ConceptualModel</code> if it is of the same instance, and throws
   * an exception otherwise.
   *
   * @param model  <code>Model</code> to be combined with this <code>ConceptualModel</code>
   * @throws IncompatibleModelExceptioin if model is not an <code>ConceptualModel</code>
   */
  public void addModel(Model mergingModel) throws IncompatibleModelException {
    if (mergingModel instanceof ConceptualModel)
      super.addModel(mergingModel);
    else
      throw new IncompatibleModelException("The model being added is not a Concept Model");
  }

}
