/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model.concept;

import belvedere.model.Constituent;
import belvedere.model.Relation;
import java.awt.*;
import java.awt.geom.*;

/**
 * Used to create ConceptualRelation objects
 */
public class ConceptualRelation extends Relation {

  //maximum length of associated text to display on the screen.
  private static final int MAX_CHAR_DISPLAY = 10;  

  //default strength of the relation, which also indicates display line thickness;
  private static final int DEFAULT_STRENGTH = 3;

  //default color for display of line and arrow
  private static final Color DEFAULT_BORDER_COLOR = new Color(0, 200, 0);

  //color for displaying arrow when selected
  private static final Color SELECTED_ARROW_COLOR = Color.yellow;

  //default color for display of text
  private static final Color DEFAULT_TEXT_COLOR = new Color(0, 0, 200);

  //default length to display arrow 
  private static final int DEFAULT_ARROW_LENGTH = 20;

  //default width to display arrow 
  private static final int DEFAULT_ARROW_WIDTH = 12;

  //default offset of arrow from the center of the line
  private static final int DEFAULT_ARROW_OFFSET = 10;

  //used to keep track of the drawing order of relations between two nodes,
  //only the first relation should draw the line
  private int relationCount;

  // The following will be used if we need to change the angle of text
  //  private AffineTransform transform;  // Transform used to draw angled text
  //  private AffineTransform save;       // Transform of original Graphics object

  private String displayString;       // String segment to be displayed graphically
  private Polygon arrow;
  private boolean associatedText;
  private Color textColor;  // color to dislay text
  private int arrowOffset = DEFAULT_ARROW_OFFSET;
  private int arrowWidth = DEFAULT_ARROW_WIDTH;
  private int arrowLength = DEFAULT_ARROW_LENGTH;

  // Testing use of sin and cos for display and rotation of arrow
  int x[] = {-10, 0, 10};
  
  /**
   * Basic constructor
   */
  public ConceptualRelation() {
    super("","","","","",Constituent.NEUTRALLY_BELIEVE);
    this.associatedText = false;
    //set up regular visual properties.
    this.regBorderColor = this.DEFAULT_BORDER_COLOR;    
    //set up visual properties to display when selected.
    this.selBorderColor = SELECTED_ARROW_COLOR;
    //set current display as regular.
    this.borderColor = this.regBorderColor;
    this.borderStroke = this.regBorderStroke;
    this.arrow = new Polygon();  // create a default polygon
    //set color to display text
    this.textColor = this.DEFAULT_TEXT_COLOR;
    this.relationCount = 0;  // default = only relation between this source+destination
  }

  /**
   * Constructor for ConceptualNode
   * 
   * @param text     text for the node
   * @param URL      URL for the node
   * @param notes    notes for the node
   * @param creator  creator of the node
   * @param modifier modifier of the node
   * @param strength strength of the node, indicated in line width
   */
  public ConceptualRelation(String text, String URL, String notes, 
                            String creator, String modifier, int strength) {
    super(text,URL,notes,creator,modifier,strength);
    if (text.length() < 10)
      this.displayString = text;
    else
      this.displayString = text.substring(0, 6) + "...";
    // check whether there is associated text with this edge
    this.associatedText = !(text.equals("")); 
    //set up regular visual properties.
    this.regBorderColor = this.DEFAULT_BORDER_COLOR;    
    //set up visual properties to display when selected.
    this.selBorderColor = SELECTED_ARROW_COLOR;
    //set current display as regular.
    this.borderColor = this.regBorderColor;
    this.borderStroke = this.regBorderStroke;
    this.arrow = new Polygon();  // create a default polygon
    //set color to display text
    this.textColor = this.DEFAULT_TEXT_COLOR;
    this.relationCount = 0;  // default = only relation between this source+destination
  }


    /**
   * Returns a new <code>Relation</code> object which is a copy of this instance.
   *
   * @return clone  a copy <code>Object</code> instance of this 
   *               <code>Relation</code>.
   */
  public Object clone() {
    return new ConceptualRelation(this.text, this.URL, this.notes, this.creator, 
                                  this.modifier, this.strength);
  }

  /**
   * This method sets the text to be displayed and calls super to set the full 
   * text associated with this object.
   *
   * @param text  Text to be associated with 
   */
  public void setText(String text) {
    super.setText(text);
    if (text.length() < 10)
      this.displayString = text;
    else
      this.displayString = text.substring(0, 6) + "...";

    this.associatedText = !(text.equals("")); 

  }

  /**
   * Paints the graphical representation of this relation, for the 
   * GraphView of Belvedere.  
   *
   * @param g           the graphics to paint the line to
   */
  public void paint(Graphics g) {
    int x1 = this.origin.getCenterX();
    int y1 = this.origin.getCenterY();
    int x2 = this.destination.getCenterX();
    int y2 = this.destination.getCenterY();
    Graphics2D g2D = (Graphics2D)g;
    g2D.setStroke(borderStroke);
    // only draw the line if it is the first to be drawn
    if (relationCount==0) {
      // the line is always in the regular color because the line may be shared
      g2D.setColor(regBorderColor);
      g2D.drawLine(x1, y1, x2, y2);
    }
    g2D.setColor(borderColor);
    paintArrow(g2D, x1, y1, x2, y2);
  }

  /**
   * Paints the graphical representation of this relation, for the 
   * GraphView of Belvedere.  
   *
   * @param g           the graphics to paint the line to
   */
  public void paint(Graphics g, int x2, int y2) {
    int x1 = this.origin.getCenterX();
    int y1 = this.origin.getCenterY();
    Graphics2D g2D = (Graphics2D)g;
    g2D.setStroke(borderStroke);
    // only draw the line if it is the first to be drawn
    if (relationCount==0) {
      // the line is always in the regular color because the line may be shared
      g2D.setColor(regBorderColor);
      g2D.drawLine(x1, y1, x2, y2);
    }
    g2D.setColor(borderColor);
    paintArrow(g2D, x1, y1, x2, y2);
  }

  /**
   * Paints a directed arrow from source to destination containing 
   * partial associated text for this relation (up to MAX_CHAR_DIPLAY).
   * The arrow is displayed relative to the source and destination points.
   *
   * @param g           the graphics2D to paint the line to
   * @param sourceX     x coordinate of one end point of line
   * @param sourceY     y coordinate of one end point of line
   * @param destX       x coordinate of second end point of line
   * @param destY       y coordinate of second end point of line
   */
  private void paintArrow(Graphics2D g, int sourceX, int sourceY, 
                          int destX, int destY) {
    int lengthX = Math.abs(sourceX - destX);
    int lengthY = Math.abs(sourceY-destY);
    int radius= (int)(3.0/4.0 * Math.sqrt((lengthX * lengthX) + (lengthY * lengthY)));
    int arrowX[] = new int[3];   // 3 x-coordinates of points 1, 2 and 3
    int arrowY[] = new int[3];   // 3 y-coordinates of points 1, 2 and 3
    double radians, cos, sin, angle; 
    arrowY[0] = 40 - radius;
    arrowY[1] = 0 - radius;
    arrowY[2] = 40 - radius;

    angle = getAngle(sourceX, sourceY, destX, destY);
    radians = (angle * Math.PI) / 180;


    cos = Math.cos(radians);
    sin = Math.sin(radians);

    // adjust the x and y coordinates of the arrow
    for (int i = 0; i < 3; i++) {
      arrowX[i] = (int)(sourceX + x[i] * cos - arrowY[i] * sin);
      arrowY[i] = (int)(sourceY + x[i] * sin + arrowY[i] * cos);    
    }

    // create a new triangle with correct visual orientation 
    this.arrow = new Polygon(arrowX, arrowY, 3);

    if (!this.associatedText) {
      // draw a hollow arrow if no associated text
      // originally using getBackground(), but it returned a different color.
      g.setColor(Color.white);
      g.fill(arrow);
      // paint arrow border to the screen
      g.setColor(this.borderColor);
      g.draw(arrow);
    } else {       // paint solid arrow and associated text to the screen
      g.setColor(this.borderColor);
      g.fill(arrow);
      g.setColor(this.textColor);
      g.drawString(displayString, arrowX[0], arrowY[0]);
    
    }
  }

  /** 
   * Returns the angle of the line between the origin and destination
   * points to which a relation must be drawn.
   *
   * @param sourceX  x-coordinate of the source
   * @param sourceY  y-coordinate of the source
   * @param destX    x-coordinate of the destination
   * @param destY    y-coordinate of the destination
   * @return angle  angle of the line representing the relation
   */
  private double getAngle(int sourceX, int sourceY, int destX, int destY) {
    final int q1orq4=0, q2orq3=2; //quadrant
    int quadrant;
    double dx = destX-sourceX;
    double dy = destY-sourceY;
    double angle;

    if (dx != 0.0)
      angle = Math.atan(dy/dx);
    else {
      if (dy >= 0.0)
        angle = Math.PI/2.0;
      else
        angle = -Math.PI/2.0;
    }
    angle = (180.0/Math.PI) * angle;
    if (dx >= 0.0 && dy >= 0.0)
      quadrant = q1orq4;
    else if (dx < 0.0)
      quadrant = q2orq3;
    else
      quadrant = q1orq4;
    switch (quadrant) {
    case q1orq4: 
      angle = angle + 90.0;
      break;
    case q2orq3: 
      angle = angle + 270.0;
      break;
    }
    return angle;
  }

  /**
   * Checks to see if point is within arrow boundaries, overwrites the 
   * default <code>belvedere.model.Relation.withinBounds</code> method.
   * This method is specific to the arrow drawn and does not relate to 
   * the line drawn between source and destination, since there might 
   * be multiple relations between the same source and destination nodes.
   *
   * @param x1     x coordinate of one end point of line
   * @param y1     y coordinate of one end point of line
   * @param x2     x coordinate of second end point of line
   * @param y2     y coordinate of second end point of line
   * @param x      x coordinate that we are testing
   * @param y      y coordinate that we are testing
   * @return true if x,y is withinBounds of the arrow, false otherwise
   */
  public boolean withinBounds(int x1, int y1, int x2, int y2, int x, int y) {
    return arrow.contains(x, y);
  }

  /**
   * returns the number of this relation, in the case of the coneptual model,
   * a relation is either the 0th or the 1st.  The relation count is used to
   * determine who should draw the line
   * 
   * @return 
   */
  public int getRelCount() {
    return this.relationCount;
  }

  /**
   * Sets the relation count
   * 
   * @param c      relation count for this relation
   */
  public void setRelCount(int c) {
    this.relationCount = c;
  }

}
