/**
 * Belvedere: tools for collaborative knowledge mapping 
 * 
 * Copyright (C) 2001, 2002 Dan Suthers
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package belvedere.model.evidence;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import javax.swing.JFrame;
import java.awt.Point;

import belvedere.model.*;
import belvedere.graphview.VisualRelation;

/**
 * abstract class foundation for evidence model relations
 */
public abstract class EvidenceRelation extends Relation {

  protected static final int FONT_SIZE = 28;

  // the monospace font makes it easier to hit the center of the circle
  protected static final Font font = new Font("Courier",Font.BOLD,FONT_SIZE);

  private boolean associatedText;

  //default length to display arrow 
  private static final int DEFAULT_CIRCLE_DIAMETER = 30;

  /**
   * Constructor for EvidenceRelation
   * 
   * @param text     text for the relation
   * @param URL      URL for the relation
   * @param notes    notes for the relation
   * @param creator  creator of the relation
   * @param modifier modifier of the relation
   * @param strength strength of the relation, indicated in line width
   */
  public EvidenceRelation(String text, String URL, String notes, 
                          String creator, String modifier, int strength) {
    super(text,URL,notes,creator,modifier,strength);
    // check whether there is associated text with this edge
    this.associatedText = !(text.equals("")); 
  }

  /**
   * This method sets the text and appropriately changes associatedText
   *
   * @param text  Text to be associated with 
   */
  public void setText(String text) {
    super.setText(text);
    this.associatedText = !(text.equals("")); 
  }

  /**
   * The function to paint the line from x1,y1 to x2, y2
   * 
   * @param g      the graphics to paint the line to
   */
  public void paint(Graphics g, String c) {    
    // paint the line
    super.paint(g);
    int x1 = this.origin.getCenterX();
    int y1 = this.origin.getCenterY();
    int x2 = this.destination.getCenterX();
    int y2 = this.destination.getCenterY();
    int midX = (x1 + x2)/2 - DEFAULT_CIRCLE_DIAMETER / 2;
    int midY = (y1 + y2)/2 - DEFAULT_CIRCLE_DIAMETER / 2;
    Graphics g2D = (Graphics2D)g;
    if (!associatedText) {
      // draw a hollow circle
      g2D.setColor(Color.white);
      g2D.fillOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
      // paint circle border to the screen
      g.setColor(this.borderColor);
      g2D.drawOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
    } else {       // paint solid circle
      g2D.setColor(this.borderColor);
      g2D.fillOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
    }
    g2D.setColor(Color.black);
    g2D.setFont(this.font);
    // the -8 and +9 should not be hard coded, we will have to determine
    // these from the font later
    g2D.drawString(c,((x1+x2)/2)-(8),((y1+y2)/2)+(9));
  }

  // paint for dragging new line state
  public void paint(Graphics g, int x2, int y2, String c) {
    super.paint(g,x2,y2);
    int x1 = this.origin.getCenterX();
    int y1 = this.origin.getCenterY();
    int midX = (x1 + x2)/2 - DEFAULT_CIRCLE_DIAMETER / 2;
    int midY = (y1 + y2)/2 - DEFAULT_CIRCLE_DIAMETER / 2;
    Graphics g2D = (Graphics2D)g;
    if (!associatedText) {
      // draw a hollow circle
      g2D.setColor(Color.white);
      g2D.fillOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
      // paint circle border to the screen
      g.setColor(this.borderColor);
      g2D.drawOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
    } else {       // paint solid circle
      g2D.setColor(this.borderColor);
      g2D.fillOval(midX,midY,DEFAULT_CIRCLE_DIAMETER,DEFAULT_CIRCLE_DIAMETER);
    }
    g2D.setColor(Color.black);
    g2D.setFont(this.font);
    // the -8 and +9 should not be hard coded, we will have to determine
    // these from the font later
    g2D.drawString(c,((x1+x2)/2)-(8),((y1+y2)/2)+(9));
  }

  /**
   * from this method a constituent will display its property editor
   * 
   * @param parent the parent component that a dialog may use to center over
   */
  public void showPropEditor(JFrame parent, Model mdl) {
    EvidenceConstituentPropertyEditor p = new EvidenceConstituentPropertyEditor(parent,"Edit Properties",this);
    p.showEditor(mdl);
  }

}
